from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType

    from tbc_video_export.common.enums import PipeType, ProcessName, TBCType


class Pipe(ABC):
    """Abstract class for pipes."""

    def __init__(self, process_name: ProcessName, tbc_type: TBCType) -> None:
        self._process_name = process_name
        self._tbc_type = tbc_type

        logging.getLogger("console").debug(
            f"Pipe ({self.pipe_type}) created for {process_name} ({tbc_type})"
        )

    @abstractmethod
    async def __aenter__(self) -> Pipe:
        """Create named pipes."""

    @abstractmethod
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None | bool:
        """Cleanup pipes and any temp dirs."""

    def __str__(self) -> str:
        """Return formatted string containing pipe paths."""
        return (
            f"{self.pipe_type}:\n"
            f"  in_path:\t{self.in_path}\n"
            f"  out_path:\t{self.out_path}\n"
            f"  in_handle:\t{self.in_handle}\n"
            f"  out_handle:\t{self.out_handle}\n"
        )

    @cached_property
    @abstractmethod
    def pipe_type(self) -> PipeType:
        """Get pipe type."""

    @cached_property
    @abstractmethod
    def in_path(self) -> Path | str:
        """Get pipe stdin string."""

    @cached_property
    @abstractmethod
    def out_path(self) -> Path | str:
        """Get pipe stdout string."""

    @property
    @abstractmethod
    def in_handle(self) -> int | None:
        """Get stdin pipe."""

    @property
    @abstractmethod
    def out_handle(self) -> int | None:
        """Get stdout pipe."""

    @abstractmethod
    def close(self) -> None:
        """Close pipe."""


@dataclass
class ConsumablePipe:
    """Class for grouping pipes together for processes."""

    tbc_type: TBCType
    consumer: ProcessName
    pipe: Pipe


# generics for wrappers
PipeInputGeneric = TypeVar("PipeInputGeneric", None, Pipe, tuple[Pipe, ...])
PipeOutputGeneric = TypeVar("PipeOutputGeneric", None, Pipe, tuple[Pipe, ...])



from __future__ import annotations

import random
import string
from datetime import datetime

from tbc_video_export.common import consts
from tbc_video_export.common.utils import ansi


def random_characters(length: int) -> str:
    """Generate N random characters from ascii and digits."""
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def current_timestamp() -> str:
    """Return the current timestamp formatted."""
    return formatted_timestamp(datetime.now())


def formatted_timestamp(ts: datetime) -> str:
    """Return a timestamp formatted."""
    return ts.strftime("%H:%M:%S:%f")[:-3]


def application_header() -> str:
    """Return the application header containing the name and version."""
    return f"{ansi.bold(consts.APPLICATION_NAME)} {consts.PROJECT_VERSION}"



from __future__ import annotations

import asyncio
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from tbc_video_export.common.enums import PipeType, ProcessName, TBCType
from tbc_video_export.common.utils import FlatList
from tbc_video_export.process.wrapper.wrapper import Wrapper

if TYPE_CHECKING:
    from tbc_video_export.process.wrapper.wrapper_config import WrapperConfig
    from tbc_video_export.program_state import ProgramState


class WrapperLDProcessVBI(Wrapper):
    """Wrapper for ld-process-vbi."""

    def __init__(self, state: ProgramState, config: WrapperConfig[None, None]) -> None:
        self._tbc_json_vbi = Path(f"{state.file_helper.input_name}.vbi.json")

        super().__init__(state, config)
        self._config = config

    def post_fn(self) -> None:  # noqa: D102
        # if dry run, just update the file name
        if self._state.dry_run:
            self._state.file_helper.tbc_json.file_name = self._tbc_json_vbi
        elif Path(self._tbc_json_vbi).is_file():
            # load the new tbc json
            self._state.file_helper.tbc_json = self._tbc_json_vbi

    @property
    def command(self) -> FlatList:  # noqa: D102
        return FlatList(
            (
                self.binary,
                "-t",
                str(self._state.opts.threads),
                "--input-json",
                self._state.file_helper.tbc_json.file_name,
                "--output-json",
                self._tbc_json_vbi,
                self._get_tbc(),
            ),
        )

    def _get_tbc(self) -> Path:
        return (
            self._state.file_helper.tbcs[TBCType.LUMA]
            if TBCType.LUMA in self._state.file_helper.tbcs
            else self._state.file_helper.tbcs[TBCType.COMBINED]
        )

    @cached_property
    def process_name(self) -> ProcessName:  # noqa: D102
        return ProcessName.LD_PROCESS_VBI

    @cached_property
    def supported_pipe_types(self) -> PipeType:  # noqa: D102
        return PipeType.NONE

    @cached_property
    def stdin(self) -> int | None:  # noqa: D102
        return None

    @cached_property
    def stdout(self) -> int | None:  # noqa: D102
        # process-vbi logs to stdout
        return asyncio.subprocess.PIPE

    @cached_property
    def stderr(self) -> int | None:  # noqa: D102
        # forward any errors to stdout
        return asyncio.subprocess.STDOUT

    @cached_property
    def log_output(self) -> bool:  # noqa: D102
        return True

    @cached_property
    def log_stdout(self) -> bool:  # noqa: D102
        return True

    @cached_property
    def env(self) -> dict[str, str] | None:  # noqa: D102
        return None

    @cached_property
    def ignore_error(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def stop_on_last_alive(self) -> bool:  # noqa: D102
        return False



from __future__ import annotations

import argparse
from typing import TYPE_CHECKING

from tbc_video_export.common import consts
from tbc_video_export.common.enums import (
    HardwareAccelType,
    VideoBitDepthType,
    VideoFormatType,
)
from tbc_video_export.opts import opt_actions, opt_types

if TYPE_CHECKING:
    from tbc_video_export.config import Config


def add_profile_opts(config: Config, parent: argparse.ArgumentParser) -> None:  # noqa: C901
    """Add profile opts to the parent arg parser."""
    # chroma/combined profiles
    profile_default = config.get_default_profile().name

    # ffmpeg arguments
    profile_opts = parent.add_argument_group("profile")

    profile_opts.add_argument(
        "--profile",
        type=str,
        choices=config.get_profile_names(),
        default=profile_default,
        metavar="profile_name",
        help="Specify an FFmpeg profile to use. "
        f"(default: {profile_default})\n"
        "See --list-profiles to see the available profiles.\n"
        "Note: These are also accessible directly, e.g. --x264"
        "\n\n",
    )

    profile_opts.add_argument(
        "--list-profiles",
        action=opt_actions.ActionListProfiles,
        config=config,
        help="Show available profiles.\n\n"
        f"You can view this in the browser here:\n"
        f"{consts.PROJECT_URL_WIKI_PROFILES}\n\n",
    )

    profile_opts.add_argument(
        "--profile-container",
        "--container",
        type=str,
        metavar="profile_container",
        help="Override an FFmpeg profile to use a specific container. Compatibility \n"
        "with profile is not guaranteed."
        "\n\n",
    )

    profile_opts.add_argument(
        "--profile-add-filter",
        "--add-filter",
        dest="profile_additional_filters",
        action="append",
        default=[],
        type=opt_types.TypeAdditionalFilter(config),
        metavar="filter_name",
        help="Use an additional filter profile when encoding. Compatibility \n"
        "with profile is not guaranteed.\n"
        "You can use this option muiltiple times."
        "\n\n",
    )

    profile_opts.add_argument(
        "--append-video-filter",
        "--append-vf",
        type=str,
        metavar="filter",
        help="Add a custom filter to the video segment of the complex filter.\n"
        "Compatibility with profile is not guaranteed.\n"
        "Use --dry-run to ensure your filter looks correct before encoding.\n\n"
        "Examples:\n"
        '--append-video-filter "scale=3480x2160:flags=lanczos,setdar=4/3"'
        "\n\n",
    )

    profile_opts.add_argument(
        "--append-other-filter",
        "--append-of",
        type=str,
        metavar="filter",
        help="Add a custom filter to the end of the complex filter.\n"
        "Compatibility with profile is not guaranteed.\n"
        "Use --dry-run to ensure your filter looks correct before encoding.\n\n"
        "Examples:\n"
        '--append-other-filter "[2:a]loudnorm=i=-14"'
        "\n\n",
    )

    # add aliases
    # hw accel aliases
    hwaccel_opts = parent.add_argument_group("hwaccel opts aliases")
    for hwaccel in HardwareAccelType:
        hwaccel_opts.add_argument(
            f"--{hwaccel.value}",
            dest="hwaccel_type",
            action=opt_actions.ActionSetVideoHardwareAccelType,
            help=argparse.SUPPRESS,
        )

    # video bit depth alias
    video_bitdepth_opts = parent.add_argument_group("video bitdepth aliases")
    for bitdepth in VideoBitDepthType:
        video_bitdepth_opts.add_argument(
            f"--{bitdepth.value}",
            dest="video_bitdepth",
            action=opt_actions.ActionSetVideoBitDepthType,
            help=argparse.SUPPRESS,
        )

    # video format aliases
    video_format_opts = parent.add_argument_group("video format aliases")

    for video_format in VideoFormatType:
        video_format_opts.add_argument(
            f"--{video_format.name.lower()}",
            dest="video_format",
            action=opt_actions.ActionSetVideoFormatType,
            help=argparse.SUPPRESS,
        )

    # profile aliases
    profile_opts = parent.add_argument_group("profile alises")

    for profile_name in config.get_profile_names():
        profile_opts.add_argument(
            f"--{profile_name}",
            default=profile_default,
            dest="profile_name",
            action=opt_actions.ActionSetProfile,
            help=argparse.SUPPRESS,
        )

    # audio profile aliases
    audio_type_opts = parent.add_argument_group("audio profile alises")

    for audio_type in config.get_audio_profile_names():
        audio_type_opts.add_argument(
            f"--{audio_type.replace('_', '-')}",
            type=str,
            action=opt_actions.ActionSetAudioOverride,
            help=argparse.SUPPRESS,
        )



from __future__ import annotations

from tbc_video_export.process.wrapper.wrapper import Wrapper
from tbc_video_export.process.wrapper.wrapper_config import (
    WrapperConfig,
)
from tbc_video_export.process.wrapper.wrapper_group import WrapperGroup

__all__ = ["Wrapper", "WrapperConfig", "WrapperGroup"]



from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common import consts, exceptions
from tbc_video_export.common.enums import (
    ChromaDecoder,
    ExportMode,
    FlagHelper,
    TBCType,
    VideoSystem,
)
from tbc_video_export.common.utils import ansi
from tbc_video_export.config.config import GetProfileFilter
from tbc_video_export.process.parser.export_state import ExportState

if TYPE_CHECKING:
    from tbc_video_export.common.file_helper import FileHelper
    from tbc_video_export.config.config import Config
    from tbc_video_export.config.profile import Profile
    from tbc_video_export.opts import Opts


@dataclass
class ProgramState:
    """Stores the program state."""

    opts: Opts
    config: Config
    file_helper: FileHelper
    export = ExportState()

    @property
    def current_export_mode(self) -> ExportMode:
        """Return the current export mode.

        This will return export_mode unless it has been set manually.
        """
        if not getattr(self, "_current_export_mode", False):
            return self.export_mode

        return self._current_export_mode

    @current_export_mode.setter
    def current_export_mode(self, export_mode: ExportMode) -> None:
        """Set the current export mode."""
        logging.getLogger("console").debug(
            f"Current export mode changed from {self.current_export_mode.name} "
            f"to {export_mode.name}."
        )
        self._current_export_mode = export_mode

    @cached_property
    def export_mode(self) -> ExportMode:
        """The export mode of the program.

        This determines the processes that will run and their configuration.
        """
        export_mode: ExportMode
        tbc_types = self.tbc_types

        # set export mode
        match tbc_types:
            case TBCType.LUMA:
                export_mode = ExportMode.LUMA
            case _ as tbc_types if TBCType.CHROMA in tbc_types:
                export_mode = (
                    ExportMode.LUMA if self.opts.luma_only else ExportMode.CHROMA_MERGE
                )
            case TBCType.COMBINED:
                export_mode = (
                    ExportMode.LUMA_EXTRACTED
                    if self.opts.luma_only
                    else ExportMode.CHROMA_COMBINED
                )
            case _:
                raise exceptions.TBCTypeError(
                    f"No export mode found for tbc type {tbc_types}."
                )

        # overwrite export mode if opts set
        if self.opts.luma_4fsc:
            export_mode = ExportMode.LUMA_4FSC

        return export_mode

    @cached_property
    def tbc_types(self) -> TBCType:
        """List of TBC types detected."""
        return self.file_helper.tbc_types

    @cached_property
    def video_system(self) -> VideoSystem:
        """Video system detected."""
        return (
            self.file_helper.tbc_json.video_system
            if self.opts.video_system is None
            else self.opts.video_system
        )

    @cached_property
    def decoder_luma(self) -> ChromaDecoder:
        """Chroma decoder for Luma TBCs."""
        # default
        decoder = ChromaDecoder.MONO

        if self.opts.chroma_decoder_luma is not None:
            decoder = self.opts.chroma_decoder_luma

        return decoder

    @cached_property
    def decoder_chroma(self) -> ChromaDecoder:
        """Chroma decoder for Chroma/Combined TBCs."""
        # default
        decoder = ChromaDecoder.TRANSFORM2D

        if self.opts.chroma_decoder is None:
            match self.video_system:
                case VideoSystem.PAL | VideoSystem.PAL_M:
                    # pal chroma
                    if TBCType.CHROMA in self.tbc_types:
                        decoder = ChromaDecoder.TRANSFORM2D

                    # pal combined
                    if self.tbc_types == TBCType.COMBINED:
                        decoder = ChromaDecoder.TRANSFORM3D

                case VideoSystem.NTSC:
                    # ntsc chroma
                    if TBCType.CHROMA in self.tbc_types:
                        decoder = ChromaDecoder.NTSC2D

                    # ntsc combined
                    if self.tbc_types == TBCType.COMBINED:
                        # check if ld or cvbs
                        decoder = (
                            ChromaDecoder.NTSC2D
                            if self.file_helper.is_combined_ld
                            else ChromaDecoder.NTSC3D
                        )
        else:
            decoder = self.opts.chroma_decoder

        return decoder

    @cached_property
    def dry_run(self) -> bool:
        """Whether the program will execute the procs or just print them."""
        return self.opts.dry_run

    @property
    def profile(self) -> Profile:
        """Return selected profile."""
        return self.config.get_profile(
            GetProfileFilter(
                self.opts.profile,
                self.opts.hwaccel_type,
                self.video_system,
            )
        )

    @cached_property
    def total_frames(self) -> int:
        """Total frames detected from the TBC json."""
        tbc_frame_count = self.file_helper.tbc_json.frame_count
        start = self.opts.start if self.opts.start is not None else 0
        length = (
            self.opts.length
            if self.opts.length is not None
            else tbc_frame_count - start
        )

        return min(tbc_frame_count - start, max(0, length))

    def __str__(self) -> str:
        """Return formatted string of program state."""
        log_files: list[str] = []

        if self.opts.log_process_output:
            log_files.append(f"{consts.CURRENT_TIMESTAMP}_*.log")

        if self.opts.debug:
            log_files.append(f"{consts.CURRENT_TIMESTAMP}_debug.log")

        log_files_str = "Disabled" if not log_files else ", ".join(log_files)

        match self.current_export_mode:
            case ExportMode.CHROMA_MERGE:
                decoders = f"{self.decoder_luma} + {self.decoder_chroma}"

            case ExportMode.CHROMA_COMBINED:
                decoders = f"{self.decoder_chroma}"

            case ExportMode.LUMA_4FSC:
                decoders = f"{TBCType.NONE}"

            case ExportMode.LUMA | ExportMode.LUMA_EXTRACTED:
                decoders = f"{self.decoder_luma}"

        if all(
            t in FlagHelper.get_flags(self.tbc_types)
            for t in (TBCType.LUMA, TBCType.CHROMA)
        ):
            tbc_type = "S-Video (Y+C)"
        else:
            tbc_type = "Composite (CVBS)"

        two_step_mode_str = "(two-step)" if self.opts.two_step else ""

        output_file: list[str] = []
        profile: list[str] = []

        if self.opts.two_step:
            output_file.append(str(self.file_helper.output_video_file_luma))

        output_file.append(str(self.file_helper.output_video_file))
        profile.append(f"{self.profile.name}")

        output_files = ", ".join(output_file)
        profiles = (
            f"{', '.join(profile)} "
            f"{'[external]' if self.config.get_config_file() is not None else ''}"
        )

        col_w: dict[str, int] = {
            "k1": 32,
            "v1": 7,
            "k2": 31,
            "v2": 17,
            "k3": 33,
            "v3": 50,
        }

        return (
            f"{ansi.dim('Input TBC:'):<{col_w['k1']}s} "
            f"{self.file_helper.tbc_luma}\n"
            f"{ansi.dim('Output Files:'):<{col_w['k1']}s} "
            f"{output_files}\n"
            f"{ansi.dim('Log Files:'):<{col_w['k1']}s} "
            f"{log_files_str}\n\n"
            f"{ansi.dim('Video System:'):<{col_w['k1']}s} "
            f"{str(self.video_system).upper():<{col_w['v1']}s}"
            f"{ansi.dim('TBC Type:'):<{col_w['k2']}s} "
            f"{tbc_type:<{col_w['v2']}s}"
            f"{ansi.dim('Chroma Decoder:'):<{col_w['k3']}s} "
            f"{decoders:<{col_w['v3']}s}\n"
            f"{ansi.dim('Total Fields:'):<{col_w['k1']}s} "
            f"{self.file_helper.tbc_json.field_count:<{col_w['v1']}d}"
            f"{ansi.dim('Total Frames:'):<{col_w['k2']}s} "
            f"{self.total_frames:<{col_w['v2']}d}"
            f"{ansi.dim('Export Mode:'):<{col_w['k3']}s} "
            f"{self.export_mode} {two_step_mode_str:<{col_w['v3']}s}\n\n"
            f"{ansi.dim('Profile:'):<{col_w['k1']}s} "
            f"{profiles}\n"
            f"{ansi.dim('Frame Type:'):<{col_w['k1']}s} "
            f"{self.opts.field_order}\n\n"
        )



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime

from tbc_video_export.common.enums import ProcessName


@dataclass
class ExportState:
    """Data from procs that represent the export state."""

    size: float = 0.0
    bitrate: str = "n/a"
    duration: str = "n/a   "
    messages: list[ExportStateMessage] = field(default_factory=list)
    concealments: int | None = None

    def merge_snapshot(self, snapshot: ExportStateSnapshot) -> None:
        """Merge a snapshot with the current export state."""
        if snapshot.size is not None:
            self.size = snapshot.size

        if snapshot.bitrate is not None:
            self.bitrate = snapshot.bitrate

        if snapshot.duration is not None:
            self.duration = snapshot.duration

        if snapshot.message is not None:
            self.messages.append(snapshot.message)

        if snapshot.concealments is not None:
            self.concealments = snapshot.concealments

    def append_message(self, message: str) -> None:
        """Append a message to the list.

        This is used when appending messages to the list from sources other than
        snapshots.
        """
        self.messages.append(
            ExportStateMessage(message, datetime.now(), ProcessName.NONE)
        )


@dataclass
class ExportStateSnapshot:
    """Snapshot data from procs, these values are merged with the export state."""

    size: float | None = None
    bitrate: str | None = None
    duration: str | None = None
    message: ExportStateMessage | None = None
    concealments: int | None = None


@dataclass
class ExportStateMessage:
    """Message data from process/wrappers."""

    message: str
    timestamp: datetime
    process: ProcessName



from __future__ import annotations

import importlib.metadata

_metadata = importlib.metadata.metadata("tbc-video-export")


def get_url_from_metadata(name: str) -> str:
    """Returns a URL from the tool.poetry.urls entry in pyproject.toml."""
    return next(
        (
            url.split(" ")[1]
            for url in _metadata.get_all("Project-URL")  # pyright: ignore[reportOptionalIterable]
            if str(url).startswith(name)
        ),
        f"{name}_url",
    )



from __future__ import annotations

from collections import abc
from collections.abc import Generator, Sequence
from pathlib import Path
from typing import TypeAlias


class FlatList:
    """List wrapper to allow adding various types into a flat list."""

    def __init__(self, values: _FlatListValues = None) -> None:
        self.data: list[str] = []
        self.append(values)

    def __str__(self) -> str:
        """Return data as a space separated string."""
        return " ".join(self.data)

    def __bool__(self) -> bool:
        """Return True contains data."""
        return len(self.data) > 0

    def append(self, values: _FlatListValues) -> None:
        """Append data to the list."""
        match values:
            case None:
                pass

            case list() | tuple() | abc.Generator():
                for v in values:
                    self.append(v)

            case FlatList():
                self.append(values.data)

            case v if v is not Sequence:
                self.data.append(str(values))

            case _:
                pass


# accepted FlatList values
_FlatListValues: TypeAlias = (
    str
    | Path
    | int
    | float
    | FlatList
    | Generator["_FlatListValues", None, None]
    | Sequence["_FlatListValues"]
    | None
)



from __future__ import annotations

from tbc_video_export.common.file_helper import FileHelper
from tbc_video_export.common.tbc_json_helper import TBCJsonHelper

__all__ = ["FileHelper", "TBCJsonHelper"]



from __future__ import annotations

import os
from dataclasses import astuple, dataclass
from typing import TYPE_CHECKING

from tbc_video_export.common.enums import PipeType
from tbc_video_export.process.wrapper.pipe.pipe_dummy import PipeDummy
from tbc_video_export.process.wrapper.pipe.pipe_os import PipeOS

if TYPE_CHECKING:
    from pathlib import Path

    from tbc_video_export.common.enums import ProcessName, TBCType
    from tbc_video_export.process.wrapper.pipe.pipe import Pipe

if os.name == "nt":
    from tbc_video_export.process.wrapper.pipe.pipe_named_nt import (
        PipeNamedNT,
    )
    from tbc_video_export.process.wrapper.pipe.pipe_named_nt_async import (
        PipeNamedNTAsync,
    )


@dataclass
class PipeFactoryConfig:
    """Config class for PipeFactory creation."""

    pipe_type: PipeType
    process_name: ProcessName
    tbc_type: TBCType
    force_dummy: bool = False
    async_nt_pipes: bool = False


class PipeFactory:
    """Factory class for Pipes."""

    @classmethod
    def create(cls, config: PipeFactoryConfig) -> Pipe:
        """Create a pipe based on PipeType."""
        pipe_type, process_name, tbc_type, force_dummy, async_nt_pipes = astuple(config)

        match pipe_type:
            # if pipe type is NULL or force null
            case pipe_type if pipe_type is PipeType.NULL or force_dummy:
                # set some readable strings for dry-runs
                in_name = f"[PIPE_IN_{str(process_name).upper()}_{tbc_type}]"
                out_name = f"[PIPE_OUT_{str(process_name).upper()}_{tbc_type}]"
                return cls.create_dummy_pipe(in_name, out_name)

            case PipeType.NAMED:
                if os.name == "posix":
                    from tbc_video_export.process.wrapper.pipe.pipe_named_posix import (
                        PipeNamedPosix,
                    )

                    return PipeNamedPosix(process_name, tbc_type)

                if os.name == "nt":
                    if async_nt_pipes:
                        return PipeNamedNTAsync()
                    return PipeNamedNT(process_name, tbc_type)

                raise NotImplementedError(f"Named pipes not implemented for {os.name}")

            case PipeType.OS:
                return PipeOS(process_name, tbc_type)

            case _:
                raise NotImplementedError(f"Could not create pipe of type {pipe_type}.")

    @classmethod
    def create_dummy_pipe(
        cls, stdin_str: Path | str = "PIPE_IN", stdout_str: Path | str = "PIPE_OUT"
    ) -> Pipe:
        """Force create a null pipe."""
        return PipeDummy(stdin_str, stdout_str)



from __future__ import annotations

import argparse
import logging
from typing import TYPE_CHECKING

from tbc_video_export.common import consts
from tbc_video_export.common.enums import (
    HardwareAccelType,
    VideoBitDepthType,
    VideoFormatType,
)
from tbc_video_export.common.utils import ansi
from tbc_video_export.config.config import GetProfileFilter

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from tbc_video_export.config import Config


class ActionDumpConfig(argparse.Action):
    """Dump configuraiton and exit action."""

    def __init__(self, config: Config, nargs: int = 0, **kwargs: Any) -> None:
        self._config = config
        super().__init__(nargs=nargs, **kwargs)

    def __call__(  # noqa: D102
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,  # noqa: ARG002
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_strings: str,  # noqa: ARG002
        *_: Any,
    ) -> None:
        self._config.dump_default_config(consts.EXPORT_CONFIG_FILE_NAME)
        parser.exit()


class ActionSetVerbosity(argparse.Action):
    """Set verbosity levels based on opts."""

    def __init__(self, nargs: int = 0, **kwargs: Any) -> None:
        super().__init__(nargs=nargs, **kwargs)

    def __call__(  # noqa: D102
        self,
        parser: argparse.ArgumentParser,  # noqa: ARG002
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_strings: str,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        if option_strings in {"--quiet", "-q"}:
            namespace.quiet = True
            namespace.no_progress = True
            namespace.show_process_output = False

        if option_strings in ("--debug"):
            namespace.debug = True
            namespace.no_progress = True

        if option_strings in ("--show-process-output"):
            namespace.show_process_output = True
            namespace.no_progress = True


class ActionListProfiles(argparse.Action):
    """Custom action for listing profiles.

    This exits the application after use.
    """

    def __init__(self, config: Config, nargs: int = 0, **kwargs: Any) -> None:
        self._config = config
        self._profile_names = config.get_profile_names()
        self._profiles_filters = config.filter_profiles
        super().__init__(nargs=nargs, **kwargs)

    def __call__(  # noqa: D102
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,  # noqa: ARG002
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_strings: str,  # noqa: ARG002
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        self._print_profiles()
        parser.exit()

    def _print_profiles(self) -> None:
        logging.getLogger("console").info(ansi.underlined("Profiles\n"))

        for profile_name in self._profile_names:
            profile = self._config.get_profile(GetProfileFilter(profile_name))
            video_profiles = self._config.get_video_profiles_for_profile(profile_name)

            data = (
                f"--{ansi.bold(profile.name)} "
                f"{'(default)' if profile.is_default else ''}\n"
            )

            for vp in video_profiles:
                data += f"{vp}\n"

            if profile.audio_profile is not None:
                data += str(profile.audio_profile)

            if profile.include_vbi:
                data += f"  {ansi.dim('Include VBI')}\t{profile.include_vbi}\n"

            logging.getLogger("console").info(data)


class ActionSetVideoHardwareAccelType(argparse.Action):
    """Set video format type with alias opts."""

    def __init__(self, nargs: int = 0, **kwargs: Any) -> None:
        super().__init__(nargs=nargs, **kwargs)

    def __call__(  # noqa: D102
        self,
        parser: argparse.ArgumentParser,  # noqa: ARG002
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_strings: str,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        # no need to check errors here, as option_strings can only be
        # VideoBitDepthType values
        namespace.hwaccel_type = HardwareAccelType(option_strings[2:].lower())


class ActionSetVideoBitDepthType(argparse.Action):
    """Set video format type with alias opts."""

    def __init__(self, nargs: int = 0, **kwargs: Any) -> None:
        super().__init__(nargs=nargs, **kwargs)

    def __call__(  # noqa: D102
        self,
        parser: argparse.ArgumentParser,  # noqa: ARG002
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_strings: str,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        # no need to check errors here, as option_strings can only be
        # VideoBitDepthType values
        match VideoBitDepthType(option_strings[2:].lower()):
            case VideoBitDepthType.BIT8:
                namespace.video_bitdepth = 8

            case VideoBitDepthType.BIT10:
                namespace.video_bitdepth = 10

            case VideoBitDepthType.BIT16:
                namespace.video_bitdepth = 16


class ActionSetVideoFormatType(argparse.Action):
    """Set video format type with alias opts."""

    def __init__(self, nargs: int = 0, **kwargs: Any) -> None:
        super().__init__(nargs=nargs, **kwargs)

    def __call__(  # noqa: D102
        self,
        parser: argparse.ArgumentParser,  # noqa: ARG002
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_strings: str,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        for format_type in VideoFormatType:
            if format_type.name.lower() == option_strings[2:].lower():
                namespace.video_format = format_type


class ActionSetProfile(argparse.Action):
    """Set profile with alias opts."""

    def __init__(self, nargs: int = 0, **kwargs: Any) -> None:
        super().__init__(nargs=nargs, **kwargs)

    def __call__(  # noqa: D102
        self,
        parser: argparse.ArgumentParser,  # noqa: ARG002
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_strings: str,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        # no need to check errors here, as option_strings can only be
        # valid profile names
        namespace.profile = option_strings[2:].lower()

class ActionSetAudioOverride(argparse.Action):
    """Set audio profile override type with alias opts."""

    def __init__(self, nargs: int = 0, **kwargs: Any) -> None:
        super().__init__(nargs=nargs, **kwargs)

    def __call__(  # noqa: D102
        self,
        parser: argparse.ArgumentParser,  # noqa: ARG002
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_strings: str,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        namespace.audio_profile = option_strings[2:].replace("-", "_").lower()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tbc_video_export.config.json import JsonConfig

DEFAULT_CONFIG: JsonConfig = {
    "profiles": [
        {
            "name": "ffv1",
            "default": True,
            "video_profile": "ffv1",
            "audio_profile": "flac",
        },
        {
            "name": "prores_hq",
            "video_profile": [
                "prores_422_hq",
                "prores_422_hq_videotoolbox",
            ],
            "audio_profile": "pcm_24",
        },
        {
            "name": "prores_4444xq",
            "video_profile": [
                "prores_4444_xq",
                "prores_4444_xq_videotoolbox",
            ],
            "audio_profile": "pcm_24",
        },
        {
            "name": "v210",
            "video_profile": "v210",
            "audio_profile": "pcm_24",
        },
        {
            "name": "v410",
            "video_profile": "v410",
            "audio_profile": "pcm_24",
        },
        {
            "name": "d10",
            "video_profile": [
                "d10_mpeg2_pal",
                "d10_mpeg2_ntsc",
            ],
            "include_vbi": True,
            "audio_profile": "pcm_24",
        },
        {
            "name": "x264",
            "video_profile": [
                "x264",
                "x264_vaapi",
                "x264_nvenc",
                "x264_quicksync",
                "x264_amf",
                "x264_videotoolbox",
            ],
            "audio_profile": "aac",
        },
        {
            "name": "x264_lossless",
            "video_profile": "x264_lossless",
            "audio_profile": "aac",
        },
        {
            "name": "x264_web",
            "video_profile": [
                "x264",
                "x264_vaapi",
                "x264_nvenc",
                "x264_quicksync",
                "x264_amf",
                "x264_videotoolbox",
            ],
            "audio_profile": "aac",
            "filter_profiles": [
                "bwdif",
            ],
        },
        {
            "name": "x265",
            "video_profile": [
                "x265",
                "x265_vaapi",
                "x265_nvenc",
                "x265_quicksync",
                "x265_amf",
                "x265_videotoolbox",
            ],
            "audio_profile": "aac",
        },
        {
            "name": "x265_lossless",
            "video_profile": "x265_lossless",
            "audio_profile": "aac",
        },
        {
            "name": "x265_web",
            "video_profile": [
                "x265",
                "x265_vaapi",
                "x265_nvenc",
                "x265_quicksync",
                "x265_amf",
                "x265_videotoolbox",
            ],
            "audio_profile": "aac",
            "filter_profiles": [
                "bwdif",
            ],
        },
        {
            "name": "av1",
            "video_profile": [
                "av1",
            ],
            "audio_profile": "aac",
        },
        {
            "name": "av1_lossless",
            "video_profile": "av1_lossless",
            "audio_profile": "aac",
        },
        {
            "name": "av1_web",
            "video_profile": [
                "av1",
            ],
            "audio_profile": "aac",
            "filter_profiles": [
                "bwdif",
            ],
        },
    ],
    "video_profiles": [
        {
            "name": "ffv1",
            "description": "FFV1",
            "codec": "ffv1",
            "video_format": "yuv422p10le",
            "container": "mkv",
            "opts": [
                "-coder",
                1,
                "-level",
                3,
                "-context",
                1,
                "-slices",
                24,
                "-slicecrc",
                1,
                "-g",
                1,
            ],
        },
        {
            "name": "prores_422_hq",
            "description": "ProRes 422 HQ",
            "codec": "prores",
            "video_format": "yuv422p10le",
            "container": "mov",
            "opts": [
                "-profile:v",
                "hq",
                "-vendor",
                "apl0",
            ],
        },
        {
            "name": "prores_422_hq_videotoolbox",
            "description": "ProRes 422 HQ - Apple Video Toolbox",
            "codec": "prores_videotoolbox",
            "video_format": "uyvy422",
            "container": "mov",
            "opts": [
                "-profile:v",
                "hq",
            ],
            "hardware_accel": "videotoolbox",
        },
        {
            "name": "prores_4444_xq",
            "description": "ProRes 4444 XQ",
            "codec": "prores",
            "video_format": "yuv444p10le",
            "container": "mov",
            "opts": [
                "-profile:v",
                "xq",
                "-vendor",
                "apl0",
            ],
        },
        {
            "name": "prores_4444_xq_videotoolbox",
            "description": "ProRes 4444 XQ - Apple Video Toolbox",
            "codec": "prores_videotoolbox",
            "video_format": "p416le",
            "container": "mov",
            "opts": [
                "-profile:v",
                "xq",
            ],
            "hardware_accel": "videotoolbox",
        },
        {
            "name": "v210",
            "description": "V210",
            "codec": "v210",
            "video_format": "yuv422p10le",
            "container": "mov",
        },
        {
            "name": "v410",
            "description": "V410",
            "codec": "v410",
            "video_format": "yuv422p10le",
            "container": "mov",
        },
        {
            "name": "d10_mpeg2_pal",
            "description": "D10 (Sony IMX/XDCAM)",
            "codec": "mpeg2video",
            "video_format": "yuv422p",
            "container": "mxf",
            "output_format": "mxf_d10",
            "opts": [
                "-minrate",
                "50000000",
                "-maxrate",
                "50000000",
                "-b:v",
                "50000000",
                "-g",
                1,
                "-flags",
                "ildct+ilme+low_delay",
                "-intra_vlc",
                1,
                "-non_linear_quant",
                1,
                "-dc",
                10,
                "-ps",
                1,
                "-qmin",
                1,
                "-qmax",
                3,
                "-bufsize",
                "2000000",
                "-rc_init_occupancy",
                "2000000",
            ],
            "video_system": "pal",
            "filter_profiles_additions": [
                "resize_pal_d10",
            ],
        },
        {
            "name": "d10_mpeg2_ntsc",
            "description": "D10 (Sony IMX/XDCAM) *BROKEN*",
            "codec": "mpeg2video",
            "video_format": "yuv422p",
            "container": "mxf",
            "output_format": "mxf_d10",
            "opts": [
                "-minrate",
                "49999840",
                "-maxrate",
                "49999840",
                "-b:v",
                "49999840",
                "-g",
                1,
                "-flags",
                "ildct+ilme+low_delay",
                "-intra_vlc",
                1,
                "-non_linear_quant",
                1,
                "-dc",
                10,
                "-ps",
                1,
                "-qmin",
                1,
                "-qmax",
                3,
                "-bufsize",
                "1668328",
                "-rc_init_occupancy",
                "1668328",
            ],
            "video_system": "ntsc",
            "filter_profiles_additions": [
                "resize_ntsc_d10",
            ],
        },
        {
            "name": "x264",
            "description": "H.264 AVC",
            "codec": "libx264",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-crf",
                18,
                "-preset",
                "slow",
                "-x264opts",
                "interlaced=1",
                "-movflags",
                "+faststart",
            ],
            "filter_profiles_override": [],
        },
        {
            "name": "x264_lossless",
            "description": "H.264 AVC - Lossless",
            "codec": "libx264",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-qp",
                0,
                "-preset",
                "veryslow",
            ],
            "filter_profiles_override": [],
        },
        {
            "name": "x264_vaapi",
            "description": "H.264 AVC - VA-API",
            "codec": "h264_vaapi",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-rc_mode",
                "CQP",
                "-global_quality",
                24,
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "vaapi",
        },
        {
            "name": "x264_nvenc",
            "description": "H.264 AVC - Nvidia NVENC - Web",
            "codec": "h264_nvenc",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-rc",
                "constqp",
                "-qp",
                22,
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "nvenc",
        },
        {
            "name": "x264_quicksync",
            "description": "H.264 AVC - Intel QuickSync - Web",
            "codec": "h264_qsv",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-global_quality",
                19,
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "quicksync",
        },
        {
            "name": "x264_amf",
            "description": "H.264 AVC - AMD AMF - Web",
            "codec": "h264_amf",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-quality",
                2,
                "-rc",
                "cqp",
                "-qp_i",
                28,
                "-qp_p",
                28,
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "amf",
        },
        {
            "name": "x264_videotoolbox",
            "description": "H.264 AVC - Apple Video Toolbox - Web",
            "codec": "h264_videotoolbox",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-profile:v",
                "main",
                "-q:v",
                60,
                "-tag:v",
                "hvc1",
            ],
            "hardware_accel": "videotoolbox",
        },
        {
            "name": "x265",
            "description": "H.265 HEVC",
            "codec": "libx265",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-crf",
                23,
                "-preset",
                "slow",
                "-movflags",
                "+faststart",
                "-x265-params",
                "interlace=true",
            ],
        },
        {
            "name": "x265_lossless",
            "description": "H.265 HEVC - Lossless",
            "codec": "libx265",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-crf",
                20,
                "-preset",
                "slow",
                "-movflags",
                "+faststart",
                "-x265-params",
                "lossless=1",
            ],
        },
        {
            "name": "x265_vaapi",
            "description": "H.265 HEVC - VA-API - Web",
            "codec": "hevc_vaapi",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-rc_mode",
                "CQP",
                "-global_quality",
                24,
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "vaapi",
        },
        {
            "name": "x265_nvenc",
            "description": "H.265 HEVC - Nvidia NVENC - Web",
            "codec": "hevc_nvenc",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-rc",
                "constqp",
                "-qp",
                24,
                "-b_ref_mode",
                0,
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "nvenc",
        },
        {
            "name": "x265_quicksync",
            "description": "H.265 HEVC - Intel QuickSync - Web",
            "codec": "hevc_qsv",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-global_quality",
                19,
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "quicksync",
        },
        {
            "name": "x265_amf",
            "description": "H.265 HEVC - AMD AMF - Web",
            "codec": "hevc_amf",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-quality",
                2,
                "-rc",
                "cqp",
                "-qp_i",
                28,
                "-qp_p",
                28,
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "amf",
        },
        {
            "name": "x265_videotoolbox",
            "description": "H.265 HEVC - Apple Video Toolbox - Web",
            "codec": "hevc_videotoolbox",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-profile:v",
                "main",
                "-q:v",
                60,
                "-tag:v",
                "hvc1",
                "-movflags",
                "+faststart",
            ],
            "hardware_accel": "videotoolbox",
        },
        {
            "name": "av1",
            "description": "AV1",
            "codec": "libsvtav1",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-crf",
                24,
                "-cpu-used",
                6,
                "-row-mt",
                1,
                "-movflags",
                "+faststart",
            ],
        },
        {
            "name": "av1_lossless",
            "description": "AV1 - Lossless",
            "codec": "libaom-av1",
            "video_format": "yuv420p",
            "container": "mov",
            "opts": [
                "-crf",
                0,
                "-cpu-used",
                6,
                "-row-mt",
                1,
                "-error-resilience",
                1,
                "-aom-params",
                "lossless=1",
            ],
        },
    ],
    "audio_profiles": [
        {
            "name": "flac",
            "description": "FLAC",
            "codec": "flac",
            "opts": [
                "-compression_level",
                12,
            ],
        },
        {
            "name": "pcm_16",
            "description": "PCM 16-bits",
            "codec": "pcm_s16le",
        },
        {
            "name": "pcm_24",
            "description": "PCM 24-bits",
            "codec": "pcm_s24le",
        },
        {
            "name": "aac",
            "description": "AAC",
            "codec": "aac",
            "opts": [
                "-ar",
                48000,
                "-b:a",
                "320K",
            ],
        },
    ],
    "filter_profiles": [
        {
            "name": "bwdif",
            "description": "BWDIF",
            "video_filter": "bwdif",
        },
        {
            "name": "colorlevels32",
            "description": "Colorlevels (32)",
            "video_filter": "colorlevels=rimin=32/255:gimin=32/255:bimin=32/255",
        },
        {
            "name": "map_l_to_lr",
            "description": "Map L to L+R",
            "other_filter": "[2:a]pan=stereo|FL=FL|FR=FL",
        },
        {
            "name": "map_r_to_lr",
            "description": "Map R to L+R",
            "other_filter": "[2:a]pan=stereo|FR=FR|FL=FR",
        },
        {
            "name": "resize_pal_standard",
            "description": "Resize to PAL standard SD resoluton (720x576)",
            "video_filter": "scale=720x576:flags=lanczos,setdar=4/3",
        },
        {
            "name": "resize_ntsc_standard",
            "description": "Resize to NTSC standard SD resoluton (720x480)",
            "video_filter": "scale=720x480:flags=lanczos,setdar=4/3",
        },
        {
            "name": "resize_pal_d10",
            "description": "Resize and pad to D10 (Sony IMX/XDCAM) resoluton (720x608)",  # noqa: E501
            "video_filter": "scale=720x608:flags=lanczos",
        },
        {
            "name": "resize_ntsc_d10",
            "description": "Resize and pad to D10 (Sony IMX/XDCAM) resoluton (720x512)",  # noqa: E501
            "video_filter": "scale=720x512:flags=lanczos",
        },
    ],
}



from __future__ import annotations

import argparse
import os
from typing import TYPE_CHECKING

from tbc_video_export.common import consts
from tbc_video_export.common.enums import (
    VideoBitDepthType,
    VideoFormatType,
    VideoSystem,
)
from tbc_video_export.opts import (
    opt_actions,
    opt_types,
    opts_ffmpeg,
    opts_ldtools,
    opts_profile,
)
from tbc_video_export.opts.opts import Opts

if TYPE_CHECKING:
    from tbc_video_export.config import Config


def parse_opts(config: Config, argv: list[str]) -> tuple[argparse.ArgumentParser, Opts]:
    """Parse program opts."""
    parser = argparse.ArgumentParser(
        prog=consts.APPLICATION_NAME,
        description=consts.PROJECT_SUMMARY,
        formatter_class=argparse.RawTextHelpFormatter,
        usage=f"{consts.APPLICATION_NAME} [options] input_file [output_file]\n\n"
        f"See --help or {consts.PROJECT_URL_WIKI_COMMANDLIST}\n"
        "---",
        epilog=f"Output/profile customization:\n{_get_opt_aliases()}",
    )

    if (cpu_count := os.cpu_count()) is None:
        cpu_count = 2

    # general
    general_opts = parser.add_argument_group("general")
    general_opts.add_argument("input_file", type=str, help="Path to the TBC file.\n\n")
    general_opts.add_argument(
        "output_file",
        type=str,
        nargs="?",
        help="Path to export to. (default: input_file.X)\n"
        "If no argument is provided, the output file will be placed in the same "
        "directory as the input_file.\n"
        "Do not specify a file extension, as this is set by the profile."
        "\n\n",
    )

    general_opts.add_argument(
        "--version",
        action="version",
        version=consts.PROJECT_VERSION,
        help="Show the program's version number and exit.\n\n",
    )

    general_opts.add_argument(
        "-t",
        "--threads",
        type=int,
        default=int(cpu_count / 2),
        metavar="int",
        help="Specify the number of concurrent threads.\n\n",
    )

    general_opts.add_argument(
        "--appimage",
        type=str,
        metavar="appimage file",
        help="Run the tools from an AppImage file.\n\n",
    )

    general_opts.add_argument(
        "--two-step",
        action="store_true",
        default=False,
        help="Enables two-step mode. (default: no)\n"
        "Enabling this will export Luma to a file before merging with Chroma.\n"
        "If you have issues using named pipes, this may be useful as a fallback.\n"
        "This is generally not required, will take longer, and will require more disk "
        "space.\n"
        "Used exclusively with export modes that merge luma and chroma."
        "\n\n",
    )

    # hidden, used for tests
    general_opts.add_argument(
        "--async-nt-pipes", action="store_true", help=argparse.SUPPRESS
    )

    general_opts.add_argument(
        "--video-system",
        type=opt_types.TypeVideoSystem(parser),
        choices=list(VideoSystem),
        metavar="format",
        help="Force a video system format. (default: from input.tbc.json)\n"
        "Available formats:\n\n  " + "\n  ".join(str(e) for e in VideoSystem) + "\n\n",
    )

    general_opts.add_argument(
        "--input-tbc-json",
        type=str,
        metavar="tbc_json_file",
        help="Specify a .tbc.json file.\n"
        "This is generally not needed and should be auto-detected.\n",
    )

    general_opts.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="Set to overwrite existing video files. (default: no)\n\n",
    )

    general_opts.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Show what would be run without running. (default: no)\n"
        "This is useful for debugging and validating opts."
        "\n\n",
    )

    general_opts.add_argument(
        "--dump-default-config",
        action=opt_actions.ActionDumpConfig,
        config=config,
        default=False,
        help=f"Dump the default configuration json to "
        f"{consts.EXPORT_CONFIG_FILE_NAME}.\n\n"
        f"For help on creating a custom profile, see:\n"
        f"{consts.PROJECT_URL_WIKI_PROFILES}"
        "\n\n",
    )

    # verbosity
    verbosity_opts = parser.add_argument_group("verbosity")
    verbosity_logger = verbosity_opts.add_mutually_exclusive_group()

    verbosity_logger.add_argument(
        "-q",
        "--quiet",
        action=opt_actions.ActionSetVerbosity,
        help="Only show ERROR messages. (default: no)\n\n",
    )

    verbosity_logger.add_argument(
        "-d",
        "--debug",
        action=opt_actions.ActionSetVerbosity,
        help="Do not suppress INFO, WARNING, or DEBUG messages. (default: no)\n"
        "If progress is enabled, this will not log to the console.\n"
        "Useful for debugging issues.\n"
        "End-users do not need this unless providing logs to developers."
        "\n\n",
    )

    # hidden, used for tests
    verbosity_opts.add_argument(
        "--no-debug-log", action="store_true", help=argparse.SUPPRESS
    )

    verbosity_opts.add_argument(
        "--no-progress",
        action="store_true",
        default=False,
        help="Hide the real-time progress display. (default: no)\n\n",
    )

    verbosity_opts.add_argument(
        "--show-process-output",
        action=opt_actions.ActionSetVerbosity,
        help="Show process output. (default: no)\nThis sets --no-progress.\n\n",
    )

    verbosity_opts.add_argument(
        "--log-process-output",
        action="store_true",
        default=False,
        help="Log process output. (default: no)\n"
        "This will log all process output to separate files."
        "\n\n",
    )

    opts_ldtools.add_ldtool_opts(parser)

    # luma
    luma_opts = parser.add_argument_group("luma")

    luma_opts.add_argument(
        "--luma-only",
        action="store_true",
        default=False,
        help="Only output a luma video. (default: no)\n"
        "For Y/C-separated TBCs, this is direct.\n"
        "For combined TBCs, filtering is applied to strip the color carrier signal out."
        "\n\n",
    )

    luma_opts.add_argument(
        "--luma-4fsc",
        action="store_true",
        default=False,
        help="This uses luma data from the TBC to produce a full signal frame export. "
        "(default: no)\n"
        "For combined TBCs, this does not filter out the chroma and thus has \n"
        "crosshatching."
        "\n\n",
    )

    opts_ffmpeg.add_ffmpeg_opts(parser)
    opts_profile.add_profile_opts(config, parser)

    opts = parser.parse_intermixed_args(argv, namespace=Opts())
    return (parser, opts)


def _get_opt_aliases() -> str:
    pix_fmts = ", ".join([f"--{t.name.lower()}" for t in VideoFormatType])
    out_str = f"  Pixel Formats:\n    {pix_fmts}\n\n"

    bitdepths = ", ".join([f"--{t.value}" for t in VideoBitDepthType])
    out_str += f"  Bit Depths:\n    {bitdepths}\n\n"

    return out_str



from __future__ import annotations

import json
import logging
from contextlib import suppress
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from tbc_video_export.common import consts, exceptions
from tbc_video_export.common.utils import files
from tbc_video_export.config.default import DEFAULT_CONFIG
from tbc_video_export.config.profile import (
    Profile,
    ProfileAudio,
    ProfileFilter,
    ProfileVideo,
)

if TYPE_CHECKING:
    from tbc_video_export.common.enums import (
        HardwareAccelType,
        VideoSystem,
    )
    from tbc_video_export.config.json import JsonConfig


class Config:
    """Profile helper.

    Uses a default json if not given a valid json file.
    Currently only contains profiles but can be extended to contain other
    program configuration.
    """

    def __init__(self) -> None:
        self._data: JsonConfig
        self._additional_filters: list[str] = []

        with suppress(FileNotFoundError, PermissionError, json.JSONDecodeError):
            # attempt to load the file if it exists
            if (file_name := self.get_config_file()) is not None:
                with Path.open(file_name, mode="r", encoding="utf-8") as file:
                    self._data = json.load(file)

        # use default config
        # not going to check for decode errors on embedded json
        if not getattr(self, "_data", False):
            self._data = DEFAULT_CONFIG

        self.profiles: list[Profile] = []

        try:
            for json_profile in self._data["profiles"]:
                for profile in self._generate_profile(json_profile["name"]):
                    self.profiles.append(profile)
        except KeyError as e:
            raise exceptions.InvalidProfileError(
                "Configuration file missing required fields.", self.get_config_file()
            ) from e

    @cached_property
    def audio_profiles(self) -> list[ProfileAudio]:
        """Return list of available audio profiles."""
        try:
            return [ProfileAudio(p) for p in self._data["audio_profiles"]]
        except KeyError as e:
            raise exceptions.InvalidAudioProfileError(
                "Could not load audio profiles.", self.get_config_file()
            ) from e

    @cached_property
    def filter_profiles(self) -> list[ProfileFilter]:
        """Return list of available filter profiles."""
        try:
            return [ProfileFilter(p) for p in self._data["filter_profiles"]]
        except KeyError as e:
            raise exceptions.InvalidFilterProfileError(
                "Could not load filter profiles.", self.get_config_file()
            ) from e

    @property
    def additional_filters(self) -> list[str]:
        """Return list of additional filter profiles."""
        return self._additional_filters

    def add_additional_filter(self, filter_name: str) -> None:
        """Append a filter to use."""
        self._additional_filters.append(filter_name)

    def get_profile(self, profile_filter: GetProfileFilter) -> Profile:
        """Return a profile from a filter."""
        try:
            profile = next(
                (profile for profile in self.profiles if profile_filter.match(profile)),
                None,
            )

            if profile is None:
                err_msg = f"Could not find profile {profile_filter.name}."

                if profile_filter.hwaccel_type is not None:
                    err_msg += f" ({profile_filter.hwaccel_type.value})"

                raise exceptions.InvalidProfileError(err_msg)

            return profile
        except KeyError as e:
            raise exceptions.InvalidProfileError(
                "Could not load profiles.", self.get_config_file()
            ) from e
        except exceptions.InvalidProfileError as e:
            raise exceptions.InvalidProfileError(str(e), self.get_config_file()) from e

    def get_profile_names(self) -> list[str]:
        """Return a list of unique profile names for a given profile type."""
        return list(dict.fromkeys(profile.name for profile in self.profiles))

    def get_default_profile(self) -> Profile:
        """Return the first default profile."""
        profile = next(
            (profile for profile in self.profiles if profile.is_default), None
        )

        if profile is None:
            raise exceptions.InvalidProfileError(
                "Unable to find default profile.", self.get_config_file()
            )

        return profile

    def get_video_profiles_for_profile(self, profile_name: str) -> list[ProfileVideo]:
        """Return list of video profiles for a given profile."""
        video_profiles: list[ProfileVideo] = []

        for profile in self.profiles:
            if profile.name == profile_name:
                video_profiles.append(profile.video_profile)

        return video_profiles

    def get_audio_profile_names(self) -> list[str]:
        """Return all audio profile names.."""
        return [audio_profile.name for audio_profile in self.audio_profiles]

    def get_filter_profile(self, filter_name: str) -> ProfileFilter:
        """Return filter profile from filter name."""
        filter_profile = next(
            (
                filter_profile
                for filter_profile in self.filter_profiles
                if filter_profile.name == filter_name
            ),
            None,
        )

        if filter_profile is None:
            raise exceptions.InvalidProfileError(
                f"Unable to find filter profile {filter_name}.",
                self.get_config_file(),
            )

        return filter_profile

    @staticmethod
    def dump_default_config(file_name: Path) -> None:
        """Attempt to dump the default config to a json file."""
        try:
            if Path(file_name).is_file():
                raise exceptions.FileIOError(
                    f"Unable to create {file_name}, already exists"
                )

            with Path.open(file_name, "w", encoding="utf-8") as file:
                json.dump(DEFAULT_CONFIG, file, ensure_ascii=False, indent=4)
        except PermissionError as e:
            raise exceptions.FileIOError(
                f"Permission error writing {file_name}."
            ) from e

        logging.getLogger("console").info(f"default config dumped to {file_name}")

    @staticmethod
    def get_config_file() -> Path | None:
        """Return name of json file to load profiles from.

        Returns None if json file does not exist in PATH or script dir.
        """
        file_name_stock = consts.EXPORT_CONFIG_FILE_NAME

        # check current dir
        if (path := Path(file_name_stock)).is_file():
            return path.absolute()

        # check binary dir
        if (
            path := Path(files.get_runtime_directory().joinpath(file_name_stock))
        ).is_file():
            return path.absolute()

        return None

    def get_profile_filters(self, profile: Profile) -> tuple[list[str], list[str]]:
        """Adds profile filters to list opts."""
        video_filters: list[str] = []
        other_filters: list[str] = []

        filter_profiles = profile.filter_profiles

        # populate filters
        for vf in (profile.video_filter for profile in filter_profiles):
            if vf is not None:
                video_filters.append(vf)

        for of in (profile.other_filter for profile in filter_profiles):
            if of is not None:
                other_filters.append(of)

        # add additional video profile filters
        for name in profile.video_profile.filter_profiles_additions:
            self._add_filter(name, video_filters, other_filters)

        # add additional opt filters
        for name in self._additional_filters:
            self._add_filter(name, video_filters, other_filters)

        return video_filters, other_filters

    def _add_filter(
        self, filter_name: str, video_filters: list[str], other_filters: list[str]
    ) -> None:
        """Add ProfileFilter to lists from name."""
        filter_profile = self.get_filter_profile(filter_name)

        if (vf := filter_profile.video_filter) is not None:
            video_filters.append(vf)

        if (of := filter_profile.other_filter) is not None:
            other_filters.append(of)

    def _generate_profile(self, profile_name: str) -> list[Profile]:
        try:
            profile_data = next(
                profile
                for profile in self._data["profiles"]
                if profile["name"] == profile_name
            )

            # get video profile(s) for profile
            if isinstance(profile_data["video_profile"], list):
                video_profiles = [
                    ProfileVideo(profile_data, json_video_profile)
                    for json_video_profile in self._data["video_profiles"]
                    for video_profile_name in profile_data["video_profile"]
                    if json_video_profile["name"] == video_profile_name
                ]
            else:
                video_profiles = [
                    next(
                        ProfileVideo(profile_data, json_video_profile)
                        for json_video_profile in self._data["video_profiles"]
                        if json_video_profile["name"] == profile_data["video_profile"]
                    )
                ]

            # get audio profile for profile
            audio_profile = next(
                (
                    ProfileAudio(json_audio_profile)
                    for json_audio_profile in self._data["audio_profiles"]
                    if json_audio_profile["name"] == profile_data["audio_profile"]
                ),
                None,
            )

            # get filter profile(s) for profile
            filter_profiles = [
                ProfileFilter(json_filter_profile)
                for json_filter_profile in self._data["filter_profiles"]
                if "filter_profiles" in profile_data
                for filter_profile_name in profile_data["filter_profiles"]
                if json_filter_profile["name"] == filter_profile_name
            ]

            profiles: list[Profile] = []

            for video_profile in video_profiles:
                profile = Profile(
                    profile_data, video_profile, audio_profile, filter_profiles
                )

                # set profile overrides
                if override := video_profile.filter_profiles:
                    profile.filter_profiles = [
                        ProfileFilter(json_filter_profile)
                        for json_filter_profile in self._data["filter_profiles"]
                        for filter_profile_name in override
                        if json_filter_profile["name"] == filter_profile_name
                    ]

                profiles.append(profile)

            return profiles
        except KeyError as e:
            raise exceptions.InvalidProfileError(
                "Unable to generate profiles.", self.get_config_file()
            ) from e


@dataclass
class GetProfileFilter:
    """Container class for get profile filter params."""

    name: str
    hwaccel_type: HardwareAccelType | None = None
    video_system: VideoSystem | None = None

    def match(self, profile: Profile) -> bool:
        """Returns true if profile matches filter."""
        video_profile = profile.video_profile

        if profile.name != self.name:
            return False

        if (
            self.hwaccel_type is not None
            and video_profile.hardware_accel is not self.hwaccel_type
        ):
            return False

        if (
            self.video_system is not None and video_profile.video_system is not None
        ) and video_profile.video_system is not self.video_system:
            return False

        return True



from __future__ import annotations

import asyncio
import logging
import os
from contextlib import suppress
from typing import TYPE_CHECKING

from tbc_video_export.common import consts
from tbc_video_export.common.enums import FlagHelper, ProcessStatus
from tbc_video_export.common.utils import log
from tbc_video_export.process.parser import Parser, ParserFactory
from tbc_video_export.process.process_state import ProcessState
from tbc_video_export.process.wrapper.pipe.pipe_os import PipeOS

if TYPE_CHECKING:
    from typing import Any

    from tbc_video_export.process.wrapper import Wrapper
    from tbc_video_export.program_state import ProgramState as ProgramState


class Process:
    """Contains the subprocess, handler and state of a running process."""

    def __init__(
        self, state: ProgramState, stop_event: asyncio.Event, wrapper: Wrapper
    ) -> None:
        self._state = state
        self._stop_event = stop_event
        self.wrapper = wrapper

        self._process_state = ProcessState()
        self._process: asyncio.subprocess.Process | None = None
        self._tasks: set[asyncio.Task[Any]] = set()

        self._output_parser = ParserFactory.create(wrapper.process_name)

        self._logger_name = (
            f"{self.wrapper.process_name}_"
            f"{FlagHelper.get_flags_str(self.wrapper.tbc_type, '_')}"
        )

    async def run(self) -> ProcessState:
        """Run the process."""
        self.state.set_has_run()

        if self._state.opts.show_process_output or self._state.opts.log_process_output:
            # setup file logger
            log.setup_logger(
                self._logger_name,
                enable_console=self._state.opts.show_process_output,
                filename=str(
                    self._state.file_helper.get_log_file(
                        self.wrapper.process_name, self.wrapper.tbc_type
                    )
                )
                if self._state.opts.log_process_output and self.wrapper.log_output
                else None,
            )

        # create subprocess
        command = self.wrapper.command.data
        logging.getLogger("console").debug(f"Running: {command}")

        self._process = await asyncio.create_subprocess_exec(
            *command,
            limit=consts.PIPE_BUFFER_SIZE,
            stdin=self.wrapper.stdin,
            stdout=self.wrapper.stdout,
            stderr=self.wrapper.stderr,
            env=dict(os.environ) | self.wrapper.env
            if self.wrapper.env is not None
            else None,
        )

        # set the process output start event
        self._process_state.set_running()

        # create parse task
        self._tasks.add(asyncio.create_task(self._parse_process_output()))

        # wait for subprocess to finish
        await self._process.wait()

        if self._process.returncode is not None:
            self._process_state.returncode = self._process.returncode

        await asyncio.wait(self._tasks)

        # even though these will be handled by the ExitStack, we should
        # close os pipes now to prevent hanging procs (POSIX only?)
        for pipe in self.wrapper.pipes:
            if isinstance(pipe, PipeOS):
                pipe.close()

        self._process_state.set_stopped()
        self._set_proc_status()

        return self._process_state

    async def stop(self) -> None:
        """Kill the subprocess and set the process state."""
        logging.getLogger("console").debug(
            f"{self.wrapper.process_name} "
            f"({FlagHelper.get_flags_str(self.wrapper.tbc_type)}) Killing process"
        )

        if self._process is not None and not self._process_state.is_stopped:
            with suppress(ProcessLookupError):
                self._process.kill()
                await self._process.wait()

        await asyncio.wait(self._tasks)

        logging.getLogger("console").debug(
            f"{self.wrapper.process_name} "
            f"({FlagHelper.get_flags_str(self.wrapper.tbc_type)}) Killing tasks"
        )

    def _set_proc_status(self) -> None:
        """Parse the returncode from a process and set the state."""
        if self._process is not None:
            if self._process.returncode == 0 or self.wrapper.ignore_error:
                self._process_state.set_success()
            else:
                self._process_state.set_error()

        logging.getLogger("console").debug(
            f"{self.wrapper.process_name} "
            f"({FlagHelper.get_flags_str(self.wrapper.tbc_type)}) "
            f"stopped with status: {self._process_state.status}"
        )

    async def _parse_process_output(self) -> None:
        """Parse a line from a process. This data is read by the progress handler."""
        while (
            not self._stop_event.is_set()
            and self.log_pipe is not None
            and ProcessStatus.RUNNING in self._process_state.status
        ):
            read_task = asyncio.Task(self.log_pipe.readline())
            await asyncio.wait([read_task], timeout=0.1)

            raw_line = await read_task

            if not raw_line:
                # done
                break

            try:
                line = raw_line.decode("utf-8").strip()

                current_state = self._output_parser.parse_line(line)

                # update export state
                self._state.export.merge_snapshot(current_state)

                # send to process logger
                logging.getLogger(self._logger_name).info(f"{line}")
            except UnicodeDecodeError:
                # skip line
                continue

    @property
    def output_parser(self) -> Parser:
        """Return output parser created for process."""
        return self._output_parser

    @property
    def log_pipe(self) -> asyncio.StreamReader | None:
        """Get the pipe to parse output from.

        Some procs use STDOUT and some use STDERR. This will be determined by the
        Wrapper.
        """
        return (
            None
            if self._process is None
            else self._process.stdout
            if self.wrapper.log_stdout
            else self._process.stderr
        )

    @property
    def state(self) -> ProcessState:
        """Return the process state."""
        return self._process_state



from __future__ import annotations

import logging
import os
from contextlib import suppress
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common.enums import PipeType
from tbc_video_export.process.wrapper.pipe.pipe import Pipe

if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType


class PipeOS(Pipe):
    """OS pipe for stdin/stdout helper."""

    _stdin: int | None = None
    _stdout: int | None = None

    async def __aenter__(self) -> Pipe:
        """Enter OS pipe context."""
        logging.getLogger("console").debug("Creating os.pipe")
        self._stdin, self._stdout = os.pipe()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None | bool:
        """Exit OS pipe context."""
        self.close()

    @cached_property
    def pipe_type(self) -> PipeType:  # noqa: D102
        return PipeType.OS

    @cached_property
    def in_path(self) -> Path | str:  # noqa: D102
        return "-"

    @cached_property
    def out_path(self) -> Path | str:  # noqa: D102
        return "-"

    @property
    def in_handle(self) -> int | None:  # noqa: D102
        return self._stdin

    @property
    def out_handle(self) -> int | None:  # noqa: D102
        return self._stdout

    def close(self) -> None:
        """Close the OS pipe."""
        logging.getLogger("console").debug("Closing pipe os.pipe")

        with suppress(OSError):
            if self.in_handle is not None and self.in_handle > -1:
                os.close(self.in_handle)
                logging.getLogger("console").debug("Closed in handle")

            if self.out_handle is not None and self.out_handle > -1:
                os.close(self.out_handle)
                logging.getLogger("console").debug("Closed out handle")



from __future__ import annotations

import re
from contextlib import suppress
from typing import TYPE_CHECKING

from tbc_video_export.process.parser.export_state import ExportStateSnapshot
from tbc_video_export.process.parser.parser import Parser

if TYPE_CHECKING:
    from tbc_video_export.common.enums import ProcessName


class ParserLDProcessVBI(Parser):
    """Parser for ld-process-vbi process."""

    def __init__(self, process_name: ProcessName) -> None:
        super().__init__(process_name)

        self.tracked_value_name = "field"

    def parse_line(self, line: str) -> ExportStateSnapshot:  # noqa: D102
        state = ExportStateSnapshot()

        patterns = [
            r"Info: Using .* threads to process (.*?) fields",
            r"Info: Processing field (.*?)$",
            (
                r"Info: VBI Processing complete - (.*?) "
                r"fields in .* seconds \( (.*?) FPS \)"
            ),
            r"Warning: .*",
            r"Critical: .*",
        ]

        with suppress(ValueError):
            if (reg := re.match("|".join(patterns), line)) is not None:
                # matches but no groups, error
                if all(v is None for v in reg.groups()):
                    self.error_count += 1
                    state.message = self._create_log_line(line)
                elif (total_fields := reg.group(1)) is not None:
                    self.tracked_value_total = int(total_fields)
                elif (current_field := reg.group(2)) is not None:
                    self.tracked_value = int(current_field)
                elif (total_fields := reg.group(3)) is not None and (
                    end_fps := reg.group(4)
                ) is not None:
                    self.tracked_value = int(total_fields)
                    self.fps = float(end_fps)

        return state

    @property
    def hide_tbc_type(self) -> bool:  # noqa: D102
        return True



from __future__ import annotations

import asyncio
import logging
from contextlib import AsyncExitStack
from datetime import datetime
from functools import partial
from itertools import chain
from typing import TYPE_CHECKING

from tbc_video_export.common.enums import ExportMode, FlagHelper, ProcessName, TBCType
from tbc_video_export.common.utils import ansi, strings
from tbc_video_export.process.process import Process
from tbc_video_export.process.progress_handler import ProgressHandler
from tbc_video_export.process.wrapper import WrapperGroup

if TYPE_CHECKING:
    from typing import Any

    from tbc_video_export.process.process_state import ProcessState
    from tbc_video_export.program_state import ProgramState


class ProcessHandler:
    """Handles the creation and running of processes."""

    def __init__(self, state: ProgramState) -> None:
        self._state = state

        self._tasks: set[asyncio.Task[None]] = set()
        self._proc_tasks: set[asyncio.Task[ProcessState]] = set()
        self._procs: dict[WrapperGroup, list[Process]] = {}

        self._progress_handler: ProgressHandler | None = None

        self._start_time = datetime.now()
        self._has_run = False

        self._stop_event = asyncio.Event()
        self._user_cancellation_event = asyncio.Event()
        self._proc_error_event = asyncio.Event()

    async def run(self) -> None:
        """Start the procs."""
        self._create_wrapper_groups()

        if self._state.dry_run:
            self._print_wrappers()
            return

        await self._task_dispatcher()
        self._stop_event.set()

        await asyncio.gather(*self._tasks)

        while pending := [task for task in self._tasks if not task.done()]:
            await asyncio.wait(pending)

        if self._has_run:
            self._print_completion_message()

    async def stop(self, cancelled_by_user: bool = False) -> None:
        """Stop process handler.

        This will wait for tasks to finish.
        """
        logging.getLogger("console").debug("Stop called on Handler")

        await asyncio.wait(self._tasks)

        if cancelled_by_user:
            self._user_cancellation_event.set()

        self._stop_event.set()

    def _create_wrapper_groups(self) -> None:
        """Create wrappers and runs processes based off export mode."""
        procs: ProcessName = ProcessName.NONE
        export_mode = self._state.current_export_mode
        tbc_types = self._state.tbc_types

        create_group = partial(WrapperGroup, self._state)

        # group 1 (vbi)
        # run process vbi
        if self._state.opts.process_vbi:
            procs |= ProcessName.LD_PROCESS_VBI

            self._procs[create_group(export_mode, TBCType.NONE, procs)] = []

        procs = ProcessName.NONE

        # group 2 (standalone)
        # run export metadata & process efm
        if self._state.opts.export_metadata:
            procs |= ProcessName.LD_EXPORT_METADATA

        if self._state.opts.process_efm:
            procs |= ProcessName.LD_PROCESS_EFM

        if procs != ProcessName.NONE:
            self._procs[create_group(export_mode, TBCType.NONE, procs)] = []

        procs = ProcessName.NONE

        # group 3 (decoding/encoding)
        if export_mode != ExportMode.LUMA_4FSC:
            if not self._state.opts.no_dropout_correct:
                procs |= ProcessName.LD_DROPOUT_CORRECT

            procs |= ProcessName.LD_CHROMA_DECODER

        # we always want ffmpeg
        procs |= ProcessName.FFMPEG

        # check for luma only
        if export_mode == ExportMode.LUMA:
            # remove chroma if luma only
            tbc_types &= ~TBCType.CHROMA

        # two step enabled, create 2 separate groups for luma/chroma
        if self._state.opts.two_step and export_mode is ExportMode.CHROMA_MERGE:
            self._procs[create_group(ExportMode.LUMA, TBCType.LUMA, procs)] = []
            self._procs[create_group(export_mode, TBCType.CHROMA, procs)] = []
        else:
            self._procs[create_group(export_mode, tbc_types, procs)] = []

    async def _task_dispatcher(self) -> Any:
        """Create the tasks and pipes."""
        # check output location is valid
        self._state.file_helper.check_output_dir()
        self._state.file_helper.check_output_file()

        for group in self._procs:
            # add the procs to the groups
            self._procs[group] = [
                Process(self._state, self._stop_event, wrapper)
                for wrapper in group.wrappers
            ]

        # create progress handler task if enabled
        if not self._state.opts.no_progress:
            self._progress_handler = ProgressHandler(
                self._state, list(chain.from_iterable(self._procs.values()))
            )

            self._tasks.add(
                asyncio.create_task(self._progress_handler.print_progress_coroutine())
            )

        self._has_run = True
        await self._run_procs()

    async def _run_procs(self) -> None:
        """Run all procs.

        This will loop through process groups and run all of the procs in a group
        before moving to the next group.
        """
        for group, procs in self._procs.items():
            async with AsyncExitStack() as pipe_stack:
                # add required pipes to stack
                for pipe_group in group.consumable_pipes:
                    await pipe_stack.enter_async_context(pipe_group.pipe)

                self._state.current_export_mode = group.export_mode

                # start task killer for group
                self._tasks.add(asyncio.create_task(self._proc_killer(procs)))

                for proc in procs:
                    self._proc_tasks.add(asyncio.create_task(proc.run()))

                # check the return values of procs as they finish and ensure they
                # are valid
                for proc_state in asyncio.as_completed(self._proc_tasks):
                    if not (await proc_state).is_successful:
                        logging.getLogger("console").debug(
                            "Process not successful, exiting"
                        )

                        # process wants to exit
                        await self._exit_all()
                        return

    async def _proc_killer(self, procs: list[Process]) -> None:
        """Proc killing task.

        This watches proc states and kills if the only remaining procs are flagged with
        stop_on_last_alive.

        Some procs (ld-dropout-correct) do not support certain arguments (-s/-l) will
        run until finished. This is solved on POSIX systems by closing the pipe once the
        consumer is finished, but this does not work on NT systems. This ensures they do
        not keep running.
        """
        while not self._proc_error_event.is_set() and not self._stop_event.is_set():
            await asyncio.sleep(0.1)

            running = [proc for proc in procs if proc.state.is_running]
            has_run_count = sum(1 for proc in procs if proc.state.has_run)

            if not len(running):
                break

            if not has_run_count:
                continue

            if has_run_count == len(procs) and all(
                proc.wrapper.stop_on_last_alive for proc in running
            ):
                for proc in running:
                    logging.getLogger("console").debug(
                        f"Killing {proc.wrapper.process_name}:{proc.wrapper.tbc_type} "
                        f"as all other procs have ended"
                    )
                    await proc.stop()

                break

    def _print_wrappers(self) -> None:
        """Print wrapper details."""
        for idx, group in enumerate(self._procs):
            logging.getLogger("console").info(ansi.dim(f"Step {idx + 1}"))

            for wrapper in group.wrappers:
                # create a string showing ENV items
                env_variables = (
                    "".join(f"{k}={v} " for k, v in wrapper.env.items())
                    if wrapper.env is not None
                    else ""
                )

                tbc_type_str = (
                    f"({FlagHelper.get_flags_str(wrapper.tbc_type, '+')})"
                    if wrapper.tbc_type is not TBCType.NONE
                    else ""
                )

                logging.getLogger("console").info(
                    ansi.dim(f"{wrapper.process_name} {tbc_type_str}\n")
                    + f"{env_variables}{wrapper.command}\n"
                )

                # proc the post-fn command to show changes on the output
                wrapper.post_fn()

    def _print_completion_message(self) -> None:
        """Print completion message based on success/failure."""
        message_suffix = (
            f" at {strings.current_timestamp()[:-4]} "
            f"after {str(datetime.now() - self._start_time)[:-3]}"
        )

        # add concealments if they were returned
        if (concealments := self._state.export.concealments) is not None:
            message_suffix += f" with {concealments} dropout concealments"

        if self._proc_error_event.is_set():
            logging.getLogger("console").error(f"Export failed{message_suffix}.")

            if self._user_cancellation_event.is_set():
                logging.getLogger("console").error("Export cancelled by user.")
        else:
            logging.getLogger("console").info(
                ansi.success_color(f"\nExport finished{message_suffix}.")
            )

    async def _exit_all(self) -> None:
        """Exit all running procs."""
        self._proc_error_event.set()

        # exit running procs
        for proc in chain.from_iterable(self._procs.values()):
            if proc.state.is_running:
                logging.getLogger("console").debug(
                    f"Stopping {proc.wrapper.process_name}"
                )

                await proc.stop()



from __future__ import annotations

import re
from contextlib import suppress
from typing import TYPE_CHECKING

from tbc_video_export.process.parser.export_state import ExportStateSnapshot
from tbc_video_export.process.parser.parser import Parser

if TYPE_CHECKING:
    from tbc_video_export.common.enums import ProcessName


class ParserLDExportMetadata(Parser):
    """Parser for ld-export-metadata process.

    This tool is silent unless there is a problem.
    """

    def __init__(self, process_name: ProcessName) -> None:
        super().__init__(process_name)

    def parse_line(self, line: str) -> ExportStateSnapshot:  # noqa: D102
        state = ExportStateSnapshot()

        patterns = [
            r".*",
        ]

        with suppress(ValueError):
            # matches but no groups, error
            if (reg := re.match("|".join(patterns), line)) is not None and all(
                v is None for v in reg.groups()
            ):
                self.error_count += 1

        return state

    @property
    def hide_tbc_type(self) -> bool:  # noqa: D102
        return True



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING

from tbc_video_export.process.parser.export_state import (
    ExportStateMessage,
    ExportStateSnapshot,
)

if TYPE_CHECKING:
    from tbc_video_export.common.enums import ProcessName


class Parser(ABC):
    """Abstract class for process parsing."""

    def __init__(self, process_name: ProcessName) -> None:
        self._process_name = process_name
        self._tracked_value_total = 0
        self._tracked_value_name = ""
        self._tracked_value = 0
        self._current_fps = 0.0
        self._error_count = 0

    @abstractmethod
    def parse_line(self, line: str) -> ExportStateSnapshot:
        """Parse process output line.

        This will set the process state in this object while returning an export state
        snapshot. This snapshot is a global state and is handled by the caller.
        """

    @property
    @abstractmethod
    def hide_tbc_type(self) -> bool:
        """Return True if we should hide the TBC type for the process.

        For some this is a pointless piece of information and should be hidden.
        """

    @property
    def tracked_value_total(self) -> int:
        """Returns total count for whatever the handler is tracking."""
        return self._tracked_value_total

    @tracked_value_total.setter
    def tracked_value_total(self, value: int) -> None:
        """Set the total count for whatever the handler is tracking."""
        self._tracked_value_total = value

    @property
    def error_count(self) -> int:
        """Return the current error count read from the process."""
        return self._error_count

    @error_count.setter
    def error_count(self, value: int) -> None:
        """Set the current error count read from the process."""
        self._error_count = value

    @property
    def tracked_value_name(self) -> str:
        """Returns the name of whatever the handler is tracking."""
        return self._tracked_value_name

    @tracked_value_name.setter
    def tracked_value_name(self, value: str) -> None:
        """Set the name of whatever the handler is tracking."""
        self._tracked_value_name = value

    @property
    def current_fps(self) -> float:
        """Returns the current FPS read from the process."""
        return self._current_fps

    @current_fps.setter
    def current_fps(self, value: float) -> None:
        """Set the current FPS read from the process."""
        self._current_fps = value

    @property
    def tracked_value(self) -> int:
        """Returns the current count for whatever the handle is tracking."""
        return self._tracked_value

    @tracked_value.setter
    def tracked_value(self, value: int) -> None:
        """Set the current count for whatever the handle is tracking."""
        self._tracked_value = value

    @property
    def process_name(self) -> ProcessName:
        """Returns the process name."""
        return self._process_name

    def _create_log_line(self, line: str) -> ExportStateMessage:
        """Return a log line object."""
        return ExportStateMessage(line, datetime.now(), self.process_name)



from __future__ import annotations

from enum import Enum, Flag, auto
from functools import cache
from typing import TypeVar

T = TypeVar("T", bound=Flag)


class FlagHelper:
    """Helper utils for enum.Flags.

    Most of these are not required in python >=3.11.
    """

    @staticmethod
    def get_flags(flag: T) -> list[T]:
        """Return a list of flags contained within a variable."""
        return [f for f in flag.__class__ if f & flag == f]

    @staticmethod
    @cache
    def get_flag_names(flag: Flag) -> list[str]:
        """Return a list of flag names contained within a variable."""
        return [
            f.name for f in flag.__class__ if f & flag == f and f.name and f.value != 1
        ]

    @staticmethod
    @cache
    def get_flags_str(flag: Flag, delimiter: str = "|") -> str:
        """Return a formatted flag string.

        This contains the flag names within a variable.
        """
        return f" {delimiter} ".join(FlagHelper.get_flag_names(flag))


class TBCType(Flag):
    """TBC type flags."""

    NONE = auto()
    COMBINED = auto()
    LUMA = auto()
    CHROMA = auto()

    def __str__(self) -> str:
        """Return enum name as string."""
        return self.name


class ExportMode(Enum):
    """Export flags for the application."""

    LUMA = "Luma"
    LUMA_EXTRACTED = "Luma (extracted)"
    LUMA_4FSC = "Luma (4FSC)"
    CHROMA_MERGE = "Luma + Chroma (merged)"
    CHROMA_COMBINED = "Luma + Chroma (combined)"

    def __str__(self) -> str:
        """Return enum value as string."""
        return self.value


class VideoSystem(Enum):
    """Supported video systems."""

    PAL = "pal"
    PAL_M = "pal_m"
    NTSC = "ntsc"

    def __str__(self) -> str:
        """Return formatted enum value as string."""
        return self.value.replace("_", "-").lower()

    @classmethod
    def _missing_(cls, value: object) -> VideoSystem | None:
        """Check if formatted string is in enum."""
        if isinstance(value, str):
            for member in cls:
                if str(member.value) == str(value):
                    return member
        return None


class VideoSystemLines(Enum):
    """Video system line values.

    FFLL / LFLL / FFRL / LFRL
    """

    PAL_DEFAULT = (22, 308, 44, 620)
    PAL_FULL_VERTICAL = (2, 308, 2, 620)
    PAL_LETTERBOX = (2, 308, 118, 548)

    NTSC_DEFAULT = (20, 259, 40, 525)
    NTSC_FULL_VERTICAL = (1, 259, 2, 525)
    NTSC_LETTERBOX = (2, 308, 118, 453)

    PAL_M_DEFAULT = NTSC_DEFAULT
    PAL_M_FULL_VERTICAL = NTSC_FULL_VERTICAL


class ChromaDecoder(Enum):
    """Available chroma decoders."""

    NONE = "none"
    PAL2D = "pal2d"
    TRANSFORM2D = "transform2d"
    TRANSFORM3D = "transform3d"
    MONO = "mono"
    NTSC1D = "ntsc1d"
    NTSC2D = "ntsc2d"
    NTSC3D = "ntsc3d"
    NTSC3DNOADAPT = "ntsc3dnoadapt"

    def __str__(self) -> str:
        """Return enum name as string."""
        return self.name


class FieldOrder(Enum):
    """Available field orders."""

    AUTO = "Auto"
    TFF = "Interlaced (Top Field First)"
    BFF = "Interlaced (Bottom Field First)"
    PROG = "Progressive"

    def __str__(self) -> str:
        """Return enum value as string."""
        return self.value


class ProcessName(Flag):
    """Process names for parsing messages."""

    NONE = auto()
    LD_PROCESS_VBI = auto()
    LD_EXPORT_METADATA = auto()
    LD_PROCESS_EFM = auto()
    LD_DROPOUT_CORRECT = auto()
    LD_CHROMA_DECODER = auto()
    FFMPEG = auto()

    def __str__(self) -> str:
        """Return formatted enum name as string."""
        return self.name.replace("_", "-").lower()


class ProcessStatus(Flag):
    """Process status flags."""

    NONE = auto()
    HAS_RUN = auto()
    RUNNING = auto()
    STOPPED = auto()
    SUCCESS = auto()
    ERROR = auto()


class PipeType(Flag):
    """Pipe types for processes."""

    NONE = auto()
    NULL = auto()
    OS = auto()
    NAMED_POSIX = auto()
    NAMED_NT = auto()
    NAMED = NAMED_POSIX | NAMED_NT

    def __str__(self) -> str:
        """Return formatted enum name as string."""
        return self.name.upper()


class VideoBitDepthType(Enum):
    """Video bitdepth types for profiles."""

    BIT8 = "8bit"
    BIT10 = "10bit"
    BIT16 = "16bit"


class HardwareAccelType(Enum):
    """Hardware accel types for profiles."""

    VAAPI = "vaapi"
    NVENC = "nvenc"
    QUICKSYNC = "quicksync"
    AMF = "amf"
    VIDEOTOOLBOX = "videotoolbox"


class VideoFormatType(Enum):
    """Video format types for profiles."""

    GRAY = {
        8: "gray8",
        10: "gray16le",
        16: "gray16le",
    }
    YUV420 = {
        8: "yuv420p",
        10: "yuv420p10le",
        16: "yuv420p16le",
    }
    YUV422 = {
        8: "yuv422p",
        10: "yuv422p10le",
        16: "yuv422p16le",
    }
    YUV444 = {
        8: "yuv444p",
        10: "yuv444p10le",
        16: "yuv444p16le",
    }

    @classmethod
    def get_new_format(cls, current_format: str, new_bitdepth: int) -> str | None:
        """Return new format from current format and new bitdepth."""
        return next(
            (
                member.value.get(new_bitdepth)
                for member in cls
                for _, v in member.value.items()
                if current_format == v
            ),
            None,
        )

    @classmethod
    def get_bitdepth(cls, current_format: str) -> int | None:
        """Return bitdepth of current format."""
        return next(
            (
                k
                for member in cls
                for k, v in member.value.items()
                if current_format == v
            ),
            None,
        )



from __future__ import annotations

from tbc_video_export.process.wrapper.pipe.pipe import (
    ConsumablePipe,
    Pipe,
    PipeInputGeneric,
    PipeOutputGeneric,
)
from tbc_video_export.process.wrapper.pipe.pipe_factory import (
    PipeFactory,
    PipeFactoryConfig,
)

__all__ = [
    "ConsumablePipe",
    "Pipe",
    "PipeInputGeneric",
    "PipeOutputGeneric",
    "PipeFactory",
    "PipeFactoryConfig",
]



from __future__ import annotations

import argparse

from tbc_video_export.common.enums import ChromaDecoder
from tbc_video_export.opts import opt_types


def add_ldtool_opts(parent: argparse.ArgumentParser) -> None:
    """Add ldtool opts to the parent argparser."""
    # decoder
    decoder_opts = parent.add_argument_group("decoder")

    decoder_opts.add_argument(
        "-s",
        "--start",
        type=int,
        metavar="int",
        help="Specify the start frame number.\n\n",
    )

    decoder_opts.add_argument(
        "-l",
        "--length",
        type=int,
        metavar="int",
        help="Specify the number of frames to process.\n\n",
    )

    decoder_opts.add_argument(
        "--reverse",
        action="store_true",
        default=False,
        help="Reverse the field order to second/first.\n\n",
    )

    decoder_opts.add_argument(
        "--output-padding",
        type=int,
        metavar="int",
        help="Pad the output frame to a multiple of these many pixels.\n\n",
    )

    decoder_lines_opts = decoder_opts.add_mutually_exclusive_group()

    decoder_lines_opts.add_argument(
        "--vbi",
        action="store_true",
        default=False,
        help="Show the VBI segment in the output video.\n"
        "This uses full-vertical and applies a crop filter.\n\n",
    )

    decoder_lines_opts.add_argument(
        "--full-vertical",
        action="store_true",
        default=False,
        help="Adjust FFLL/LFLL/FFRL/LFRL for full vertical export.\n\n",
    )

    decoder_lines_opts.add_argument(
        "--letterbox",
        action="store_true",
        default=False,
        help="Adjust FFLL/LFLL/FFRL/LFRL for letterbox crop.\n\n",
    )

    decoder_opts.add_argument(
        "--first-active-field-line",
        "--ffll",
        type=int,
        metavar="int",
        help="The first visible line of a field.\n\n"
        "  Range 1-259 for NTSC (default: 20)\n"
        "        2-308 for PAL  (default: 22)"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--last-active-field-line",
        "--lfll",
        type=int,
        metavar="int",
        help="The last visible line of a field.\n\n"
        "  Range 1-259 for NTSC (default: 259)\n"
        "        2-308 for PAL  (default: 308)"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--first-active-frame-line",
        "--ffrl",
        type=int,
        metavar="int",
        help="The first visible line of a field.\n\n"
        "  Range 1-525 for NTSC (default: 40)\n"
        "        1-620 for PAL  (default: 44)"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--last-active-frame-line",
        "--lfrl",
        type=int,
        metavar="int",
        help="The last visible line of a field.\n\n"
        "  Range 1-525 for NTSC (default: 525)\n"
        "        1-620 for PAL  (default: 620)"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--chroma-decoder",
        type=opt_types.TypeChromaDecoder(parent),
        choices=list(ChromaDecoder),
        metavar="decoder",
        help="Set the chroma decoder to be used.\n"
        "Available decoders:\n\n"
        f"  {ChromaDecoder.MONO} (default for LUMA)\n\n"
        f"  {ChromaDecoder.PAL2D}\n"
        f"  {ChromaDecoder.TRANSFORM2D} (default for PAL/PAL-M S-Video)\n"
        f"  {ChromaDecoder.TRANSFORM3D} (default for PAL/PAL-M CVBS)\n\n"
        f"  {ChromaDecoder.NTSC1D}\n"
        f"  {ChromaDecoder.NTSC2D} (default for NTSC S-Video and CVBS)\n"
        f"  {ChromaDecoder.NTSC3D} (default for NTSC CVBS LD)\n"
        f"  {ChromaDecoder.NTSC3DNOADAPT}"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--chroma-decoder-luma",
        type=opt_types.TypeChromaDecoder(parent),
        choices=list(ChromaDecoder),
        metavar="decoder",
        help="Set the chroma decoder to be used for luma.\n"
        "You likely do not need to touch this.\n"
        "Available decoders:\n\n"
        f"  {ChromaDecoder.MONO} (default)\n\n"
        f"  {ChromaDecoder.PAL2D}\n"
        f"  {ChromaDecoder.TRANSFORM2D}\n"
        f"  {ChromaDecoder.TRANSFORM3D}\n\n"
        f"  {ChromaDecoder.NTSC1D}\n"
        f"  {ChromaDecoder.NTSC2D}\n"
        f"  {ChromaDecoder.NTSC3D}\n"
        f"  {ChromaDecoder.NTSC3DNOADAPT}"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--chroma-gain",
        type=float,
        metavar="float",
        help="Gain factor applied to chroma components.\n"
        "This does not apply to the luma decoder for Y/C-separated TBCs.\n"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--chroma-phase",
        type=float,
        metavar="float",
        help="Phase rotation applied to chroma components in degrees.\n"
        "This does not apply to the luma decoder for Y/C-separated TBCs.\n"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--luma-nr",
        type=float,
        metavar="float",
        help="Luma noise reduction level in dB.\n"
        "You likely do not need to use this and should instead apply any\n"
        "noise reduction in post.\n\n"
        "Note for Y/C-separated TBCs:\n"
        "  This will only apply to the luma instance.\n"
        "  Requires the luma decoder to be set to something other than mono.\n"
        "\n\n",
    )

    decoder_opts.add_argument(
        "--transform-threshold",
        type=float,
        metavar="float",
        help="Transform: Uniform similarity threshold in 'threshold' mode.\n\n",
    )

    decoder_opts.add_argument(
        "--transform-thresholds",
        type=str,
        metavar="file_name",
        help="Transform: File containing per-bin similarity thresholds in "
        "'threshold' mode."
        "\n\n",
    )

    # decoder (ntsc)
    ntsc_decoder_opts = parent.add_argument_group("decoder (NTSC)")
    ntsc_decoder_opts.add_argument(
        "--chroma-nr",
        type=float,
        metavar="float",
        help="Chroma noise reduction level in dB.\n"
        "This does not apply to the luma decoder for Y/C-separated TBCs.\n"
        "You likely do not need to use this and should instead apply any\n"
        "noise reduction in post.\n"
        "\n\n",
    )

    ntsc_decoder_opts.add_argument(
        "--ntsc-phase-comp",
        action=argparse.BooleanOptionalAction,
        help="Enable or disable adjusting phase per line using burst phase.\n"
        "S-Video and CVBS have this option enabled by default.\n"
        "CVBS LD has this option disabled by default."
        "\n\n",
    )

    ntsc_decoder_opts.add_argument(
        "--oftest",
        action="store_true",
        default=False,
        help="Overlay the adaptive filter map (only used for testing).\n\n",
    )

    # decoder (pal)
    pal_decoder_opts = parent.add_argument_group("decoder (PAL)")
    pal_decoder_opts.add_argument(
        "--simple-pal",
        action="store_true",
        default=False,
        help="Transform: Use 1D UV filter.\n"
        "Only available with Transform decoders."
        "\n\n",
    )

    # dropout-correct
    dropout_correct_opts = parent.add_argument_group("dropout correction")

    dropout_correct_opts.add_argument(
        "--no-dropout-correct",
        action="store_true",
        default=False,
        help="Disable dropout correction. (default: no)\n"
        "This will run ld-chroma-decoder without dropout correction."
        "\n\n",
    )

    # process-vbi
    process_vbi = parent.add_argument_group("process vbi")
    process_vbi.add_argument(
        "--process-vbi",
        action="store_true",
        default=False,
        help="Run ld-process-vbi before exporting. (default: no)\n\n"
        "Note: The generated JSON file will be used for decoding."
        "\n\n",
    )

    process_vbi.add_argument(
        "--process-vbi-keep-going",
        action="store_true",
        default=False,
        help="Keep going on errors. (default: no)\n\n",
    )

    # process-efm (EXPTERIMENTAL)
    process_efm = parent.add_argument_group("process efm (EXPERIMENTAL)")
    process_efm.add_argument(
        "--process-efm",
        action="store_true",
        default=False,
        help="Run ld-process-efm before exporting. (default: no)\n\n",
    )

    process_efm.add_argument(
        "--process-efm-keep-going",
        action="store_true",
        default=False,
        help="Keep going on errors. (default: no)\n\n",
    )

    process_efm.add_argument(
        "--process-efm-dts",
        action="store_true",
        default=False,
        help="Audio is DTS rather than PCM. (default: no)\n\n",
    )

    # export-metadata
    export_metadata = parent.add_argument_group("export metadata")

    export_metadata.add_argument(
        "--export-metadata",
        action="store_true",
        default=False,
        help="Run ld-export-metadata before exporting. (default: no)\n\n"
        "Note: The generated subtitles and ffmetadata will be muxed when encoding."
        "\n\n",
    )

    export_metadata.add_argument(
        "--export-metadata-keep-going",
        action="store_true",
        default=False,
        help="Keep going on errors. (default: no)\n\n",
    )



from __future__ import annotations

import logging
import os
import platform
from contextlib import contextmanager
from functools import cache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Any


@contextmanager
def create_terminal_buffer() -> Generator[None, Any, Any]:
    """Handle entering and exiting the alternative buffer."""
    try:
        # enter alternative buffer and hide cursor
        logging.getLogger("progress").info(
            f"{enable_alternative_buffer()}{hide_cursor()}"
        )

        yield
    finally:
        # exit alternative buffer and restore cursor
        logging.getLogger("progress").info(
            f"{disable_alternative_buffer()}{show_cursor()}"
        )


@cache
def has_ansi_support() -> bool:
    """Check for ANSI support in the terminal."""
    # logging.getLogger("console").debug("Activating ANSI support ")

    if os.name == "nt":
        # check if Windows >= 10.0.14393 for ansi console support.
        # This includes Windows 10 and Windows 11.
        return (
            platform.release() == "10"
            and platform.version() >= "10.0.14393"
            and os.isatty(0)
        )

    return os.isatty(0)


# style wrappers


@cache
def default_color(text: str) -> str:
    """Return text wrapped with the default color."""
    return f"{_default_color()}{text}{_reset_color()}"


@cache
def error_color(text: str) -> str:
    """Return text wrapped with error color."""
    return f"{_error_color()}{text}{_reset_color()}"


@cache
def success_color(text: str) -> str:
    """Return text wrapped with success color."""
    return f"{_success_color()}{text}{_reset_color()}"


@cache
def progress_color(text: str) -> str:
    """Return text wrapped with progress color."""
    return f"{_progress_color()}{text}{_reset_color()}"


@cache
def bold(text: str) -> str:
    """Return text wrapped with bold style."""
    return f"{_bold()}{text}{_reset_bold()}"


@cache
def italic(text: str) -> str:
    """Return text wrapped with italic style."""
    return f"{_italic()}{text}{_reset_italic()}"


@cache
def dim(text: str) -> str:
    """Return text wrapped with dim color."""
    return f"{_dim_color()}{text}{_reset_color()}"


@cache
def dim_style(text: str) -> str:
    """Return text wrapped with dim style."""
    return f"{_dim()}{text}{_reset_dim()}"


@cache
def underlined(text: str) -> str:
    """Return text wrapped with underlined style."""
    return f"{_underlined()}{text}{_reset_underlined()}"


# Erase Functions


@cache
def erase_from_cursor() -> str:
    """Return erase from cursor (to end of screen) escape code."""
    return "\x1B[0J"


@cache
def erase_line() -> str:
    """Return erase current line escape code."""
    return "\x1B[0K"


@cache
def erase_screen() -> str:
    """Return erase screen escape code."""
    return "\x1B[2J"


# Cursor Controls


@cache
def move_to_home() -> str:
    """Return move to home (0, 0) escape code."""
    return "\x1B[H"


@cache
def go_up_lines(count: int) -> str:
    """Return escape code to go up N lines."""
    return f"\x1B[{count}A"


# Screen Modes


@cache
def enable_alternative_buffer() -> str:
    """Return enable alternative buffer escape code."""
    return "\x1B[?1049h"


@cache
def disable_alternative_buffer() -> str:
    """Return disable alternative buffer escape code."""
    return "\x1B[?1049l"


@cache
def show_cursor() -> str:
    """Return show cursor escape code."""
    return "\x1B[?25h"


@cache
def hide_cursor() -> str:
    """Return hide cursor escape code."""
    return "\x1B[?25l"


# Color codes
def _default_color() -> str:
    """Return default color escape code."""
    return "\x1B[38;5;255m" if has_ansi_support() else ""  # white (255)


def _dim_color() -> str:
    """Return dim color escape code."""
    return "\x1B[38;5;245m" if has_ansi_support() else ""  # grey (240)


def _error_color() -> str:
    """Return error color escape code."""
    return "\x1B[0;31m" if has_ansi_support() else ""  # red


def _success_color() -> str:
    """Return success color escape code."""
    return "\x1B[0;32m" if has_ansi_support() else ""  # green


def _progress_color() -> str:
    """Return progress color escape code."""
    return "\x1B[0;36m" if has_ansi_support() else ""  # cyan


def _reset_color() -> str:
    """Return color reset escape code."""
    return "\x1B[0;39m" if has_ansi_support() else ""


# Colors / Graphics Mode


def _bold() -> str:
    """Return bold escape code."""
    return "\x1B[1m" if has_ansi_support() else ""


def _reset_bold() -> str:
    """Return end bold escape code."""
    return "\x1B[22m" if has_ansi_support() else ""


def _italic() -> str:
    """Return italic escape code."""
    return "\x1B[23m" if has_ansi_support() else ""


def _reset_italic() -> str:
    """Return end italic escape code."""
    return "\x1B[23m" if has_ansi_support() else ""


def _dim() -> str:
    """Return dim escape code."""
    return "\x1B[2m" if has_ansi_support() else ""


def _reset_dim() -> str:
    """Return end dim escape code."""
    return "\x1B[22m" if has_ansi_support() else ""


def _underlined() -> str:
    """Return underlined escape code."""
    return "\x1B[4m" if has_ansi_support() else ""


def _reset_underlined() -> str:
    """Return end underlined escape code."""
    return "\x1B[24m" if has_ansi_support() else ""



from __future__ import annotations

import asyncio
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common.enums import PipeType, ProcessName
from tbc_video_export.common.utils import FlatList
from tbc_video_export.process.wrapper.wrapper import Wrapper

if TYPE_CHECKING:
    from tbc_video_export.process.wrapper.wrapper_config import WrapperConfig
    from tbc_video_export.program_state import ProgramState


class WrapperLDExportMetadata(Wrapper):
    """Wrapper for ld-export-metadata."""

    def __init__(self, state: ProgramState, config: WrapperConfig[None, None]) -> None:
        super().__init__(state, config)
        self._config = config

    def post_fn(self) -> None:  # noqa: D102
        pass

    @property
    def command(self) -> FlatList:  # noqa: D102
        return FlatList(
            (
                self.binary,
                "--ffmetadata",
                self._state.file_helper.ffmetadata_file,
                "--closed-captions",
                self._state.file_helper.cc_file,
                self._state.file_helper.tbc_json.file_name,
            ),
        )

    @cached_property
    def process_name(self) -> ProcessName:  # noqa: D102
        return ProcessName.LD_EXPORT_METADATA

    @cached_property
    def supported_pipe_types(self) -> PipeType:  # noqa: D102
        return PipeType.NONE

    @cached_property
    def stdin(self) -> int | None:  # noqa: D102
        return None

    @cached_property
    def stdout(self) -> int | None:  # noqa: D102
        return asyncio.subprocess.DEVNULL

    @cached_property
    def stderr(self) -> int | None:  # noqa: D102
        return asyncio.subprocess.PIPE

    @cached_property
    def log_output(self) -> bool:  # noqa: D102
        return True

    @cached_property
    def log_stdout(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def env(self) -> dict[str, str] | None:  # noqa: D102
        return None

    @cached_property
    def ignore_error(self) -> bool:  # noqa: D102
        return False

    @cached_property
    def stop_on_last_alive(self) -> bool:  # noqa: D102
        return False



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing_extensions import NotRequired


class JsonConfig(TypedDict):
    """Raw mapping of config from JSON."""

    profiles: list[JsonProfile]
    video_profiles: list[JsonSubProfileVideo]
    audio_profiles: list[JsonSubProfileAudio]
    filter_profiles: list[JsonSubProfileFilter]


class JsonProfile(TypedDict):
    """Raw mapping of profile from JSON."""

    name: str
    default: NotRequired[bool]
    include_vbi: NotRequired[bool]
    video_profile: str | list[str]
    audio_profile: NotRequired[str]
    filter_profiles: NotRequired[list[str]]


class JsonSubProfile(TypedDict):
    """Raw mapping of subprofile from JSON."""

    name: str
    description: str


class JsonSupportedFormats(TypedDict):
    """Raw mapping of supported formats from JSON."""

    bitdepth: int
    format: str


class JsonSubProfileVideo(JsonSubProfile):
    """Raw mapping of video subprofile from JSON."""

    container: str
    output_format: NotRequired[str]
    codec: str
    opts: NotRequired[list[str | int]]
    video_system: NotRequired[str]
    video_format: str
    filter_profiles_additions: NotRequired[list[str]]
    filter_profiles_override: NotRequired[list[str]]
    hardware_accel: NotRequired[str]


class JsonSubProfileAudio(JsonSubProfile):
    """Raw mapping of audio subprofile from JSON."""

    codec: str
    opts: NotRequired[list[str | int]]


class JsonSubProfileFilter(JsonSubProfile):
    """Raw mapping of filter subprofile from JSON."""

    video_filter: NotRequired[str]
    other_filter: NotRequired[str]



from __future__ import annotations

import ctypes
import ctypes.wintypes
import logging
import os
from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, cast

from tbc_video_export.common import consts

if TYPE_CHECKING:
    from types import TracebackType
    from typing import ClassVar

assert os.name == "nt"

import win32api  # noqa: E402
import win32con  # noqa: E402

GetLastError = cast(Callable[[], int], win32api.GetLastError)  # pyright: ignore [reportUnknownMemberType]

# stubs for nt fns
kernel32 = ctypes.windll.kernel32
CreateToolhelp32Snapshot = ctypes.windll.kernel32.CreateToolhelp32Snapshot
Process32First = ctypes.windll.kernel32.Process32First
Process32Next = ctypes.windll.kernel32.Process32Next
CloseHandle = ctypes.windll.kernel32.CloseHandle
GetStdHandle = kernel32.GetStdHandle


class VirtualTerminal:
    """Context for enabling/disabling Virtual Terminal Processing on NT."""

    def __init__(self) -> None:
        self._original_console_mode: ctypes.wintypes.DWORD | None = None

    def __enter__(self) -> VirtualTerminal:
        """Enter Virtual Terminal context."""
        logging.getLogger("console").debug("Entering VirtualTerminal context")

        # store original config
        self._original_console_mode = _get_console_mode()

        # enables Virtual terminal Processing aka ANSI sequence support
        # this should work on all terminals unless Legacy Console mode is enabled
        if self._original_console_mode is not None:
            _set_console_mode(
                self._original_console_mode.value
                | consts.NT_ENABLE_VIRTUAL_TERMINAL_PROCESSING
            )

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit Virtual Terminal context."""
        logging.getLogger("console").debug("Leaving VirtualTerminal context")

        if self._original_console_mode is not None:
            _set_console_mode(self._original_console_mode)

        # Await user input if launched via explorer
        if _get_grandparent_name() == "explorer.exe":
            input("Press any key to exit.")


@staticmethod
def _get_grandparent_name() -> str | None:
    """Get name of grandparent process."""
    ppid = os.getppid()

    # get grandparent pid
    if (parent := _get_pid_data(ppid)) is not None and (
        gparent := _get_pid_data(parent.ppid)
    ) is not None:
        return gparent.exe_file
    return None


@staticmethod
def _get_pid_data(pid: int) -> _PidData | None:
    """Loops through procs and returns data for pid if found."""
    process_snapshot: ctypes.wintypes.HANDLE | None = None
    pe32 = _PROCESSENTRY32()
    pe32.dwSize = ctypes.sizeof(_PROCESSENTRY32)

    try:
        # create snapshot of procs
        process_snapshot = CreateToolhelp32Snapshot(consts.NT_TH32CS_SNAPPROCESS, 0)
        # get parent pid of parent
        if Process32First(process_snapshot, ctypes.byref(pe32)) == win32con.FALSE:
            logging.getLogger("console").debug(
                f"Process32First failed: {GetLastError()}"
            )
            return None

        # attempt to find pid in snapshot
        while Process32Next(process_snapshot, ctypes.byref(pe32)) != win32con.FALSE:
            if pe32.th32ProcessID == pid:
                return _PidData(
                    pe32.th32ProcessID,
                    pe32.th32ParentProcessID,
                    pe32.szExeFile.decode(),
                )

        logging.getLogger("console").debug(f"Process32Next failed: {GetLastError()}")
    finally:
        # return None
        if process_snapshot is not None:
            CloseHandle(process_snapshot)


@staticmethod
def _get_stdout_handle() -> int:
    """Return handle for STDOUT."""
    return GetStdHandle(consts.NT_STD_OUTPUT_HANDLE)


@staticmethod
def _set_console_mode(mode: int | ctypes.wintypes.DWORD) -> None:
    """Set the console mode."""
    if not kernel32.SetConsoleMode(_get_stdout_handle(), mode):
        logging.getLogger("console").debug(f"SetConsoleMode failed: {GetLastError()}")


@staticmethod
def _get_console_mode() -> ctypes.wintypes.DWORD | None:
    """Get the current console mode."""
    mode = ctypes.wintypes.DWORD()

    if not kernel32.GetConsoleMode(_get_stdout_handle(), ctypes.byref(mode)):
        logging.getLogger("console").debug(f"GetConsoleMode failed: {GetLastError()}")
        return None

    return mode


@dataclass
class _PidData:
    pid: int
    ppid: int
    exe_file: str


class _PROCESSENTRY32(ctypes.Structure):
    _fields_: ClassVar = [
        ("dwSize", ctypes.c_ulong),
        ("cntUsage", ctypes.c_ulong),
        ("th32ProcessID", ctypes.c_ulong),
        ("th32DefaultHeapID", ctypes.wintypes.LPARAM),
        ("th32ModuleID", ctypes.c_ulong),
        ("cntThreads", ctypes.c_ulong),
        ("th32ParentProcessID", ctypes.c_ulong),
        ("pcPriClassBase", ctypes.c_ulong),
        ("dwFlags", ctypes.c_ulong),
        ("szExeFile", ctypes.c_char * 260),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from tbc_video_export.common.enums import (
    ChromaDecoder,
    FieldOrder,
    VideoSystem,
)

if TYPE_CHECKING:
    import argparse

    from tbc_video_export.config import Config


class TypeVideoSystem:
    """Return ChromaDecoder value if it exists."""

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        self._parser = parser

    def __call__(self, value: str) -> VideoSystem:  # noqa: D102
        try:
            return VideoSystem[value.replace("-", "_").upper()]
        except KeyError:
            self._parser.error(
                f"argument --video-system: invalid VideoSystem value: '{value}', "
                f"check --help for available options."
            )


class TypeFieldOrder:
    """Return FieldOrder value if it exists."""

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        self._parser = parser

    def __call__(self, value: str) -> FieldOrder:  # noqa: D102
        try:
            return FieldOrder[value.upper()]
        except KeyError:
            self._parser.error(
                f"argument --field-order: invalid FieldOrder value: '{value}', "
                f"check --help for available options."
            )


class TypeAdditionalFilter:
    """Return ProfileFilter if it exists."""

    def __init__(self, config: Config) -> None:
        self._config = config

    def __call__(self, value: str) -> str:  # noqa: D102
        # add to config
        self._config.add_additional_filter(value)
        return value


class TypeOverrideAudioProfile:
    """Return ProfileFilter if it exists."""

    def __init__(self, config: Config) -> None:
        self._config = config

    def __call__(self, value: str) -> str:  # noqa: D102
        # add to config
        self._config.add_additional_filter(value)
        return value


class TypeChromaDecoder:
    """Return ChromaDecoder value if it exists."""

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        self._parser = parser

    def __call__(self, value: str) -> ChromaDecoder:  # noqa: D102
        try:
            return ChromaDecoder[value.upper()]
        except KeyError:
            self._parser.error(
                f"argument --chroma-decoder: invalid ChromaDecoder value: '{value}', "
                f"check --help for available options."
            )



from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING

from tbc_video_export.common import exceptions
from tbc_video_export.common.enums import (
    ExportMode,
    FlagHelper,
    PipeType,
    ProcessName,
    TBCType,
)
from tbc_video_export.process.wrapper import Wrapper, WrapperConfig
from tbc_video_export.process.wrapper.pipe import (
    ConsumablePipe,
    Pipe,
    PipeFactory,
    PipeFactoryConfig,
)
from tbc_video_export.process.wrapper.wrapper_ffmpeg import WrapperFFmpeg
from tbc_video_export.process.wrapper.wrapper_ld_chroma_decoder import (
    WrapperLDChromaDecoder,
)
from tbc_video_export.process.wrapper.wrapper_ld_dropout_correct import (
    WrapperLDDropoutCorrect,
)
from tbc_video_export.process.wrapper.wrapper_ld_export_metadata import (
    WrapperLDExportMetadata,
)
from tbc_video_export.process.wrapper.wrapper_ld_process_efm import WrapperLDProcessEFM
from tbc_video_export.process.wrapper.wrapper_ld_process_vbi import WrapperLDProcessVBI

if TYPE_CHECKING:
    from tbc_video_export.program_state import ProgramState


class WrapperGroup:
    """Create a group of wrappers.

    This is based on requested procs, tbc type and export mode.
    These groups are used to run procs in "order", as some procs must run before others.
    """

    def __init__(
        self,
        state: ProgramState,
        export_mode: ExportMode,
        tbc_types: TBCType,
        process_names: ProcessName,
    ) -> None:
        self._state = state
        self._export_mode = export_mode
        self._tbc_types = tbc_types
        self._process_names = process_names

        self.wrappers: list[Wrapper] = []
        self.consumable_pipes: list[ConsumablePipe] = []

        self._create_pipe_config = partial(
            PipeFactoryConfig,
            async_nt_pipes=self._state.opts.async_nt_pipes,
            force_dummy=self._state.dry_run,
        )

        logging.getLogger("console").debug(
            f"Creating wrappers for {FlagHelper.get_flags_str(process_names, '+')}"
        )

        self._create_standalone_wrappers()
        self._create_decoder_wrappers()
        self._create_ffmpeg_wrapper()

    @property
    def export_mode(self) -> ExportMode:
        """Return export mode for the wrapper group."""
        return self._export_mode

    def _create_standalone_wrappers(
        self,
    ) -> None:
        """Create standalone wrappers that do not rely on other procs or pipes."""
        wrapper_config = WrapperConfig[None, None](
            self._export_mode, TBCType.NONE, None, None
        )

        if ProcessName.LD_PROCESS_VBI in self._process_names:
            self.wrappers.append(WrapperLDProcessVBI(self._state, wrapper_config))

        if ProcessName.LD_PROCESS_EFM in self._process_names:
            self.wrappers.append(WrapperLDProcessEFM(self._state, wrapper_config))

        if ProcessName.LD_EXPORT_METADATA in self._process_names:
            self.wrappers.append(WrapperLDExportMetadata(self._state, wrapper_config))

    def _create_decoder_wrappers(self) -> None:
        """Create decoder wrappers.

        This optionally includes dropout correction and decoding for every tbc type.
        """
        # do not add dropout-correct & decoder on 4fsc mode
        if self._export_mode == ExportMode.LUMA_4FSC:
            return

        for tbc_type in FlagHelper.get_flags(self._tbc_types):
            create_pipe_config = partial(self._create_pipe_config, tbc_type=tbc_type)

            if ProcessName.LD_DROPOUT_CORRECT in self._process_names:
                # create dropout correction -> decoder pipe
                self.consumable_pipes.append(
                    ConsumablePipe(
                        tbc_type,
                        ProcessName.LD_CHROMA_DECODER,
                        pipe := PipeFactory.create(
                            create_pipe_config(
                                PipeType.OS, ProcessName.LD_DROPOUT_CORRECT
                            )
                        ),
                    )
                )

                # create dropout correct wrapper
                self.wrappers.append(
                    WrapperLDDropoutCorrect(
                        self._state,
                        WrapperConfig[None, Pipe](
                            self._export_mode, tbc_type, None, pipe
                        ),
                    )
                )

            if ProcessName.LD_CHROMA_DECODER in self._process_names:
                if not self._get_pipes_for_consumer(
                    ProcessName.LD_CHROMA_DECODER, tbc_type
                ):
                    # if no pipes have been created for chroma-decoder, create a
                    # dummy pipe with the tbc file name
                    self.consumable_pipes.append(
                        ConsumablePipe(
                            tbc_type,
                            ProcessName.LD_CHROMA_DECODER,
                            PipeFactory.create_dummy_pipe(
                                self._state.file_helper.tbcs[tbc_type]
                            ),
                        )
                    )

                # create decoder -> ffmpeg pipe
                self.consumable_pipes.append(
                    ConsumablePipe(
                        tbc_type,
                        ProcessName.FFMPEG,
                        pipe := PipeFactory.create(
                            create_pipe_config(
                                PipeType.OS
                                if self._state.opts.two_step
                                else PipeType.NAMED,
                                ProcessName.LD_CHROMA_DECODER,
                            )
                        ),
                    )
                )

                # create decoder wrapper
                self.wrappers.append(
                    WrapperLDChromaDecoder(
                        self._state,
                        WrapperConfig[Pipe, Pipe](
                            self._export_mode,
                            tbc_type,
                            self._get_pipe_for_consumer(
                                ProcessName.LD_CHROMA_DECODER, tbc_type
                            ),
                            pipe,
                        ),
                    )
                )

    def _create_ffmpeg_wrapper(self) -> None:
        """Create wrapper for ffmpeg process."""
        if ProcessName.FFMPEG in self._process_names:
            # check if any pipes created for ffmpeg, if not create
            # a dummy pipe  using the tbc file name
            if not self._get_pipes_for_consumer(ProcessName.FFMPEG, self._tbc_types):
                self.consumable_pipes.append(
                    ConsumablePipe(
                        self._tbc_types,
                        ProcessName.FFMPEG,
                        PipeFactory.create_dummy_pipe(self._state.file_helper.tbc_luma),
                    )
                )

            # create ffmpeg wrapper
            self.wrappers.append(
                WrapperFFmpeg(
                    self._state,
                    WrapperConfig[tuple[Pipe], None](
                        self._export_mode,
                        self._tbc_types,
                        self._get_pipes_for_consumer(
                            ProcessName.FFMPEG, self._tbc_types
                        ),
                        None,
                    ),
                )
            )

    def _get_pipes_for_consumer(
        self, consumer_name: ProcessName, tbc_types: TBCType
    ) -> tuple[Pipe, ...]:
        """Get pipes created for a consumer/wrapper."""
        pipes = tuple(
            group.pipe
            for group in self.consumable_pipes
            if group.consumer == consumer_name and group.tbc_type in tbc_types
        )
        # ensure we do not have multiple os stdio pipes
        if sum(1 for pipe in pipes if pipe.pipe_type is PipeType.OS) > 1:
            raise exceptions.PipeError(
                f"Multiple {PipeType.OS} pipes are not supported."
            )

        return pipes

    def _get_pipe_for_consumer(
        self, consumer_name: ProcessName, tbc_types: TBCType
    ) -> Pipe:
        """Get single pipe created for a consumer/wrapper."""
        return next(iter(self._get_pipes_for_consumer(consumer_name, tbc_types)))



from __future__ import annotations

import re
from contextlib import suppress
from typing import TYPE_CHECKING

from tbc_video_export.process.parser.export_state import ExportStateSnapshot
from tbc_video_export.process.parser.parser import Parser

if TYPE_CHECKING:
    from tbc_video_export.common.enums import ProcessName


class ParserLDProcessEFM(Parser):
    """Parser for ld-process-efm process."""

    def __init__(self, process_name: ProcessName) -> None:
        super().__init__(process_name)

        self.tracked_value_name = "%"
        self.tracked_value_total = 100

    def parse_line(self, line: str) -> ExportStateSnapshot:  # noqa: D102
        state = ExportStateSnapshot()

        patterns = [
            r"Info: Processed (.*?)%$",
            r"Info:      Corrupt samples: (.*?)$",
            r"Info:      Missing samples: (.*?)$",
        ]

        with suppress(ValueError):
            if (reg := re.match("|".join(patterns), line)) is not None:
                if (percentage := reg.group(1)) is not None:
                    self.tracked_value = int(percentage)
                elif (samples := reg.group(2)) is not None or (
                    samples := reg.group(3)
                ) is not None:
                    self.error_count += int(samples)

        return state

    @property
    def hide_tbc_type(self) -> bool:  # noqa: D102
        return True



from __future__ import annotations

import asyncio
import logging
import os
import queue
from dataclasses import dataclass
from enum import Flag, auto
from functools import cached_property
from typing import TYPE_CHECKING

from tbc_video_export.common import consts, exceptions
from tbc_video_export.common.enums import PipeType
from tbc_video_export.common.utils import strings
from tbc_video_export.process.wrapper.pipe.pipe import Pipe

if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType


assert os.name == "nt"

import pywintypes  # noqa: E402
import win32event  # noqa: E402
import win32file  # noqa: E402
import win32pipe  # noqa: E402
import win32security  # noqa: E402
import winerror  # noqa: E402

from tbc_video_export.common.utils import win32  # noqa: E402


class PipeNamedNTAsync(Pipe):
    """Named pipe helper for NT systems.

    This uses asynchronous I/O compared to PipedNamedNT. We do not have
    to make a thread for every blocking win32 function.

    This is currently unused and is being tested for performance and
    stability.

    I hate Windows named pipes.
    """

    def __init__(self) -> None:
        self._id = strings.random_characters(7)
        self._pipe_name = rf"\\.\pipe\{consts.APPLICATION_NAME}-{self._id}"
        self.pipe_bridge = AsyncNamedPipeBridge(
            AsyncNamedPipeBridgeConfig(self._pipe_name)
        )
        self._bridge_thread: asyncio.Future[None] | None = None

    async def __aenter__(self) -> Pipe:
        """Enter async NT pipe context."""
        self._bridge_thread = asyncio.get_event_loop().run_in_executor(
            None, self.pipe_bridge.run
        )

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None | bool:
        """Exit async NT pipe context."""
        self.close()

    @cached_property
    def pipe_type(self) -> PipeType:  # noqa: D102
        return PipeType.NULL

    @cached_property
    def in_path(self) -> Path | str:  # noqa: D102
        return self._pipe_name

    @cached_property
    def out_path(self) -> Path | str:  # noqa: D102
        return self._pipe_name

    @property
    def in_handle(self) -> int | None:  # noqa: D102
        return None

    @property
    def out_handle(self) -> int | None:  # noqa: D102
        return None

    def close(self) -> None:  # noqa: D102
        self.pipe_bridge.disconnect_all()


class PipeState(Flag):
    """The current named pipe state."""

    NONE = auto()
    CONNECTING = auto()
    READING = auto()
    WRITING = auto()
    WRITE_TO_PEERS = auto()
    PENDING = auto()


class PipeInstance:
    """Represents a named pipe."""

    def __init__(
        self,
        pipe_id: int,
        handle: int,
        pipe_buffer_size: int = 4 * 1024 * 1024,  # 4MB
    ) -> None:
        self.pipe_id = pipe_id
        self.handle = handle
        self.buffer = win32file.AllocateReadBuffer(pipe_buffer_size)
        self.state = PipeState.NONE

        self.ol = win32file.OVERLAPPED()
        self.ol.hEvent = win32event.CreateEvent(
            win32security.SECURITY_ATTRIBUTES(), True, True, None
        )
        self.ol.Offset = 0
        self.ol.OffsetHigh = 0

        self.pending_write_size: int = -1

        self.current_read_size = 0
        self.read_data: memoryview | None = None
        self.write_queue: queue.Queue[memoryview] = queue.Queue()

    def log(self, message: str) -> None:
        """Log string with pipe details prefixed."""
        logging.getLogger("console").debug(f"[{self.pipe_id}][{self.state}] {message}")


@dataclass
class AsyncNamedPipeBridgeConfig:
    """Container for async pipe bridge config."""

    pipe_name: str
    peer_count: int = 2
    minimum_peers: int = 2
    pipe_timeout: int = 10000
    event_timeout: int = 100
    pipe_buffer_size: int = 65536  # advisory, doesn't really matter
    stop_on_disconnect: bool = True


class AsyncNamedPipeBridge:
    """Handle bridging between named pipes.

    Allows N pipes to connect to a pipe and acts as an echo server between them.
    While this is only used to go from A->B, A<->N is also possible.
    """

    def __init__(self, config: AsyncNamedPipeBridgeConfig) -> None:
        self.pipes: list[PipeInstance] = []
        self._pipe_name = config.pipe_name
        self._peer_count = config.peer_count
        self._minimum_peers = config.minimum_peers
        self._pipe_timeout = config.pipe_timeout
        self._event_timeout = config.event_timeout
        self._pipe_buffer_size = config.pipe_buffer_size
        self._stop_on_disconnect = config.stop_on_disconnect

        self._stop_queued = False

        for pipe_id in range(self._peer_count):
            self.pipes.append(self._create(pipe_id))

    def _create(self, pipe_id: int) -> PipeInstance:
        """Create the pipe instance."""
        try:
            handle = win32pipe.CreateNamedPipe(  # pyright: ignore [reportUnknownMemberType]
                self._pipe_name,
                win32pipe.PIPE_ACCESS_DUPLEX | win32file.FILE_FLAG_OVERLAPPED,
                win32pipe.PIPE_READMODE_BYTE
                | win32pipe.PIPE_TYPE_BYTE
                | win32pipe.PIPE_WAIT,
                self._peer_count,
                self._pipe_buffer_size,
                self._pipe_buffer_size,
                self._pipe_timeout,
                win32security.SECURITY_ATTRIBUTES(),
            )
        except pywintypes.error:
            match win32.GetLastError():
                case winerror.ERROR_BAD_NETPATH:
                    raise exceptions.PipeNTError(
                        f"Invalid pipe path {self._pipe_name}"
                    ) from None
                case _ as e:
                    raise exceptions.PipeNTError(
                        f"Create pipe failed with: {e}"
                    ) from None
        else:
            if handle == win32file.INVALID_HANDLE_VALUE:
                raise exceptions.PipeNTError(
                    f"Create pipe failed with {win32.GetLastError()}"
                )

            # create the pipe instance
            pipe = PipeInstance(pipe_id, handle)
            self._connect(pipe)

            return pipe

    def _connect(self, pipe: PipeInstance) -> None:
        """Connect a pipe instance."""
        result = win32pipe.ConnectNamedPipe(pipe.handle, pipe.ol)  # pyright: ignore [reportUnknownMemberType, reportUnknownVariableType]
        if isinstance(result, int):
            self._handle_connect_ret(pipe, result)
        else:
            # this should never reach, we're only checking for type checker
            raise exceptions.PipeNTError(f"ConnectNamedPipe returned {result}")

    def _disconnect(self, pipe: PipeInstance, close_handle: bool = False) -> None:
        """Disconnect a pipe instance."""
        try:
            pipe.log("Disconnecting")
            win32pipe.DisconnectNamedPipe(pipe.handle)
        except pywintypes.error:
            self._handle_disconnect_ret(pipe)

        if close_handle:
            try:
                win32file.CloseHandle(pipe.handle)
            except pywintypes.error:
                self._handle_disconnect_ret(pipe)

    def disconnect_all(self) -> None:
        """Disconnect all pipe instances."""
        for pipe in (pipe for pipe in self.pipes):
            self._disconnect(pipe, True)

    def _handle_connection(self, pipe: PipeInstance):
        """Handle new/re-connection."""
        try:
            # get connect result
            ol_result = win32file.GetOverlappedResult(pipe.handle, pipe.ol, False)

            return self._handle_connect_ret(pipe, ol_result)
        except pywintypes.error:
            return self._handle_connect_ret(pipe, win32.GetLastError())

    @staticmethod
    def _handle_connect_ret(pipe: PipeInstance, ret: int) -> bool:
        """Handle connection return."""
        match ret:
            case winerror.ERROR_SUCCESS:
                pipe.log("Connection successful")

                # set the pipe state based on whether the queue already contains data
                pipe.state = (
                    PipeState.WRITING
                    if not pipe.write_queue.empty()
                    else PipeState.READING  # pyright: ignore [reportUnknownMemberType, reportUnknownArgumentType]
                )

                return True

            # The overlapped connection in progress.
            case winerror.ERROR_IO_PENDING:
                pipe.log("Connection pending")
                pipe.state = PipeState.CONNECTING

            # Client is already connected, so signal an event.
            # There is a process on other end of the pipe.
            case winerror.ERROR_PIPE_CONNECTED:
                pipe.log("Pipe already connected")
                pipe.state = PipeState.CONNECTING
                win32event.SetEvent(pipe.ol.hEvent)

            # Overlapped I/O event is not in a signaled state.
            case winerror.ERROR_IO_INCOMPLETE:
                pipe.log("Overlapped I/O event is not in a signaled state.")

            case _ as e:
                raise exceptions.PipeNTError(f"Connect failed with: {e}")

        return False

    def _handle_disconnect(self, pipe: PipeInstance) -> None:
        """Handle disconnects."""
        try:
            if self._stop_on_disconnect:
                pipe.log("Stop on disconnect set")
                self._stop_queued = True

            self._disconnect(pipe, self._stop_queued)

            if not self._stop_queued:
                self._connect(pipe)
        except pywintypes.error as e:
            pipe.log(f"Failed to disconnect {e}")

    @staticmethod
    def _handle_disconnect_ret(pipe: PipeInstance) -> None:
        """Handle disconnect return."""
        match win32.GetLastError():
            case winerror.ERROR_PIPE_NOT_CONNECTED:
                pipe.log("Attempted to disconnect disconnected pipe")

            case winerror.ERROR_INVALID_HANDLE:
                pipe.log("Attempted to close invalid handle")

            case _ as e:
                pipe.log(f"Disconnect error: {e}")
                raise SystemExit

    def _handle_read(self, pipe: PipeInstance) -> None:
        """Handle reads."""
        try:
            result, _ = win32file.ReadFile(pipe.handle, pipe.buffer, pipe.ol)
            self._handle_read_ret(pipe, result)
        except pywintypes.error:
            self._handle_read_ret(pipe, win32.GetLastError())

    def _handle_read_pending(self, pipe: PipeInstance) -> None:
        """Handle pending reads.."""
        try:
            self._handle_read_ol_ret(pipe)

            # remove pending flag
            pipe.state &= ~PipeState.PENDING
        except pywintypes.error:
            self._handle_read_ret(pipe, win32.GetLastError())

    def _handle_read_ol_ret(self, pipe: PipeInstance) -> None:
        """Handle read result.."""
        try:
            # get result
            ol_result = win32file.GetOverlappedResult(pipe.handle, pipe.ol, False)

            # this shouldn't happen
            if ol_result == 0:
                pipe.log("Read {ol_result} bytes, expected > 0")
                self._handle_disconnect(pipe)

            # keep track of current read size for memoryview splicing
            pipe.current_read_size += ol_result

            if pipe.read_data is None:
                pipe.read_data = pipe.buffer[:ol_result]  # pyright: ignore [reportGeneralTypeIssues]
            else:
                pipe.read_data[pipe.current_read_size :] = pipe.buffer[:ol_result]  # pyright: ignore [reportGeneralTypeIssues, reportUnknownArgumentType]

            # copy data to all connected peers
            for peer in (
                peer
                for peer in self.pipes
                if peer.state is not PipeState.CONNECTING
                and peer.pipe_id != pipe.pipe_id
            ):
                # add read data to other pipe instance write queue
                if isinstance(pipe.read_data, memoryview):  # pyright: ignore [reportUnknownMemberType]
                    peer.write_queue.put_nowait(pipe.read_data)
                    peer.pending_write_size = ol_result
                    peer.state = PipeState.WRITING

            # prep pipe for writing to peers
            pipe.current_read_size = 0
            pipe.read_data = None
            pipe.state = PipeState.WRITE_TO_PEERS
        except pywintypes.error:
            self._handle_read_ret(pipe, win32.GetLastError())
        except queue.Full as e:
            # queue should never be full as we prioritise writes
            # something is broke if we reach here
            raise exceptions.PipeNTError("Write queue full.") from e

    def _handle_read_ret(self, pipe: PipeInstance, ret: int) -> None:
        """Handle read returns."""
        match ret:
            case winerror.ERROR_SUCCESS:
                self._handle_read_ol_ret(pipe)

            # More data is available.
            case winerror.ERROR_MORE_DATA:
                bytes_read = len(pipe.buffer)  # pyright: ignore [reportGeneralTypeIssues]

                # keep track of current read size for memoryview splicing
                pipe.current_read_size += bytes_read

                if pipe.read_data is None:
                    pipe.read_data = pipe.buffer[:bytes_read]  # pyright: ignore [reportGeneralTypeIssues]
                else:
                    pipe.read_data[pipe.current_read_size :] = pipe.buffer[:bytes_read]  # pyright: ignore [reportGeneralTypeIssues, reportUnknownArgumentType]

                # set pending and read more
                pipe.state |= PipeState.PENDING
                win32file.ReadFile(pipe.handle, pipe.buffer, pipe.ol)

            case winerror.ERROR_IO_PENDING:
                pipe.state |= PipeState.PENDING

            case _:
                self._handle_ret(pipe, ret)

    def _handle_write(self, pipe: PipeInstance) -> None:
        """Handle writes."""
        try:
            data = pipe.write_queue.get()

            result, bytes_written = win32file.WriteFile(
                pipe.handle,
                data,  # pyright: ignore [reportGeneralTypeIssues]
                pipe.ol,
            )

            # confirm we wrote the correct amount of data
            self._handle_write_ret(
                pipe, result, bytes_written == pipe.pending_write_size
            )
        except pywintypes.error:
            self._handle_write_ret(pipe, win32.GetLastError(), False)

    def _handle_write_pending(self, pipe: PipeInstance) -> None:
        """Handle pending writes.."""
        try:
            self._handle_write_ol_ret(pipe)
        except pywintypes.error:
            self._handle_write_ret(pipe, win32.GetLastError(), False)

    @staticmethod
    def _handle_write_ol_ret(pipe: PipeInstance) -> None:
        """Handle write result."""
        # get result
        ol_result = win32file.GetOverlappedResult(pipe.handle, pipe.ol, False)

        # we wrote the correct amount of data, set pipe state to reading
        if ol_result == pipe.pending_write_size:
            pipe.state = PipeState.READING
            return

    def _handle_write_ret(
        self,
        pipe: PipeInstance,
        ret: int,
        success_condition: bool,
    ) -> None:
        """Handle write return."""
        match ret:
            case winerror.ERROR_SUCCESS:
                if success_condition:
                    pipe.state = PipeState.READING

            case winerror.ERROR_IO_PENDING:
                pipe.state |= PipeState.PENDING

            case _:
                self._handle_ret(pipe, ret)

    def _handle_writepeer(self, pipe: PipeInstance) -> None:
        """Handle writing to peers."""
        try:
            # write to all connected pipes excluding sender
            for peer in (
                peer
                for peer in self.pipes
                if peer.state is not PipeState.CONNECTING
                and peer.pipe_id != pipe.pipe_id
            ):
                self._handle_write(peer)

            pipe.state = PipeState.READING

        except pywintypes.error:
            self._handle_write_ret(pipe, win32.GetLastError(), False)

    def _handle_ret(self, pipe: PipeInstance, ret: int) -> None:
        """Handle general returns."""
        match ret:
            case (
                # The pipe has been ended.
                winerror.ERROR_BROKEN_PIPE
                # The pipe is being closed.
                | winerror.ERROR_NO_DATA
                # No process is on the other end of the pipe.
                | winerror.ERROR_PIPE_NOT_CONNECTED
            ):
                self._handle_disconnect(pipe)

            # Waiting for a process to open the other end of the pipe.
            case winerror.ERROR_PIPE_LISTENING:
                pass

            case (
                # The pipe state is invalid
                winerror.ERROR_BAD_PIPE
                # All pipe instances are busy.
                | winerror.ERROR_PIPE_BUSY
                # Overlapped I/O event is not in a signaled state.
                | winerror.ERROR_IO_INCOMPLETE
            ):
                raise exceptions.PipeNTError(f"[{pipe.pipe_id}] Pipe in invalid state.")

            case _ as e:
                raise exceptions.PipeNTError(
                    f"[{pipe.pipe_id}] An unknown error occured: {e}"
                )

    def _process(self, pipe: PipeInstance) -> None:
        """Process data based on pipe state."""
        match pipe.state:
            case PipeState.CONNECTING:
                if not self._handle_connection(pipe):
                    return

            # Pending read handler
            case _ as state if PipeState.READING | PipeState.PENDING in state:
                self._handle_read_pending(pipe)

            case _ as state if PipeState.WRITING | PipeState.PENDING in state:
                self._handle_write_pending(pipe)

            case PipeState.READING:
                self._handle_read(pipe)

            case PipeState.WRITING:
                self._handle_write(pipe)

            case PipeState.WRITE_TO_PEERS:
                self._handle_writepeer(pipe)
            case _:
                pipe.log("Shouldn't be here")
                raise exceptions.PipeNTError(
                    f"[{pipe.pipe_id}] An unknown error occured."
                )

    def run(self):
        """Run the pipe bridge."""
        while True:
            connected_peer_count = sum(
                1 for pipe in self.pipes if pipe.state is not PipeState.CONNECTING
            )

            pending_writers = [
                pipe
                for pipe in self.pipes
                if pipe.state is PipeState.WRITING | PipeState.PENDING
            ]

            # if only 1 peer, only listen for events from unconnected peers
            if connected_peer_count < self._minimum_peers:
                pipe_events = [
                    pipe for pipe in self.pipes if pipe.state is PipeState.CONNECTING
                ]
            elif count := len(pending_writers):
                # If there are pipes with pending writes, only process them
                pipe_events = [
                    pipe
                    for pipe in self.pipes
                    if pipe.state is PipeState.WRITING | PipeState.PENDING
                ]
            else:
                # Multiple peers
                pipe_events = self.pipes

            wait_idx = -1

            try:
                wait_idx = win32event.WaitForMultipleObjects(  # pyright: ignore [reportUnknownMemberType, reportUnknownVariableType]
                    [pipe.ol.hEvent for pipe in pipe_events], False, self._event_timeout
                )
            except pywintypes.error:
                match win32.GetLastError():
                    case winerror.ERROR_INVALID_PARAMETER:
                        continue

                    case _ as e:
                        logging.getLogger("console").debug(f"Unknown events error {e}")
                        return

            assert isinstance(wait_idx, int)

            if wait_idx == winerror.WAIT_TIMEOUT:
                continue

            wait_idx -= win32event.WAIT_OBJECT_0
            pipe = pipe_events[wait_idx]

            if self._stop_queued:
                logging.getLogger("console").debug(
                    "Stop queued, checking remaining pipes"
                )
                if (
                    count := sum(1 for p in self.pipes if not p.write_queue.empty())
                ) == 0:
                    logging.getLogger("console").debug(
                        "Remaining pipes are all empty, disconnecting"
                    )
                    self.disconnect_all()
                    return

                logging.getLogger("console").debug(
                    f"Waiting for {count} pipes to finish writing"
                )

            self._process(pipe)



from __future__ import annotations

import asyncio
import logging
import os
import sys
from contextlib import nullcontext

from tbc_video_export.common import FileHelper, exceptions
from tbc_video_export.common.utils import interrupts, log, strings
from tbc_video_export.config.config import Config
from tbc_video_export.opts import opt_validators, opts_parser
from tbc_video_export.process.process_handler import ProcessHandler
from tbc_video_export.program_state import ProgramState


def main(argv: list[str] = sys.argv[1:]) -> None:
    """Entry point for tbc-video-export."""
    asyncio.run(_run(argv))


async def _run(argv: list[str]) -> None:
    log.setup_logger("console")

    # set to INFO initially, opts will determine level
    logging.getLogger("console").setLevel(logging.INFO)

    asyncio.get_event_loop().set_exception_handler(exceptions.loop_exception_handler)

    try:
        config = Config()
        parser, opts = opts_parser.parse_opts(config, argv)
        log.set_verbosity(opts)

        files = FileHelper(opts, config)
        state = ProgramState(
            opts,
            config,
            files,
        )

        opt_validators.validate_opts(state, parser, opts)

        handler = ProcessHandler(state)

        logging.getLogger("console").info(
            f"{strings.application_header()}\n\n{state}\n"
        )

        if os.name == "nt":
            from tbc_video_export.common.utils import win32

            terminal_ctx = win32.VirtualTerminal()
        else:
            terminal_ctx = nullcontext()

        with interrupts.InterruptHandler(handler), terminal_ctx:
            await handler.run()
    except Exception as e:  # noqa: BLE001
        exceptions.handle_exceptions(e)
    finally:
        logging.shutdown()


if __name__ == "__main__":
    main()
